import pandas as pd
import xarray as xr
from loguru import logger
from mt_metadata.processing.aurora.decimation_level import (
    DecimationLevel as AuroraDecimationLevel,
)
from mth5.processing import KernelDataset


def extract_features(
    dec_level_config: AuroraDecimationLevel, tfk_dataset: KernelDataset
) -> pd.DataFrame:
    """
        Temporal place holder.

        In the future, features will be stored in the MTH5 and accessible as df or xr.
        Here we compute them in place -- its inefficient, but will allow as faster
        end-to-end prototype.

        Note that the container with the

        Development Note #4:
            Features that will be used for FC-weighting are expected to share a time axis with the FC data.
            (This could possibly be relaxed in the future with some interpolation, but for now, enforce it).
            This means that striding window features derived from time series should be implemented using the
            WindowedTimeSeries methods as in run_ts_to_stft (including truncation to clock-zero) etc.

    Parameters
    ----------
    dec_level_config
    tfk_dataset

    Returns
    -------

    """
    try:
        features = read_features_from_mth5()
        return features
    except Exception as e:
        msg = f"Features could not be accessed from MTH5 -- {e}\n"
        msg += "Calculating features on the fly (development only)"
        logger.info(msg)

    for (
        chws
    ) in dec_level_config.channel_weight_specs:  # This refers to solving a TF equation
        # Loop over features and compute them
        msg = f"channel weight spec:\n {chws}"
        logger.debug(msg)
        for fws in chws.feature_weight_specs:
            msg = f"feature weight spec: {fws}"
            logger.debug(msg)
            feature = fws.feature
            msg = f"feature: {feature}"
            logger.debug(msg)
            msg = f"feature type: {type(feature).__name__}, has validate_station_ids: {hasattr(feature, 'validate_station_ids')}"
            logger.debug(msg)
            feature_chunks = []
            if feature.name == "coherence":
                msg = f"{feature.name} is not supported as a data weighting feature"
                logger.warning(msg)
            elif feature.name == "multiple_coherence":
                msg = f"{feature.name} is not supported as a data weighting feature"
                logger.warning(msg)
                # raise NotImplementedError(msg)
            elif feature.name == "striding_window_coherence":
                # TODO: review logic in time_series_helpers.run_ts_to_stft.  This logic could be
                #  modified to work similarly.  Importantly, we need to map the time axis of stft
                #  onto the features
                feature.validate_station_ids(
                    local_station_id=tfk_dataset.local_station_id,
                    remote_station_id=tfk_dataset.remote_station_id,
                )

                # Make main window scheme agree with the one used for the FC calculation
                feature.window = dec_level_config.stft.window
                feature.set_subwindow_from_window(fraction=0.25)

                # get data required for computing the feature.
                # Loop the runs (or run-pairs) ... this should be equivalent to grouping on start time.
                # TODO: consider mixing in valid run info from processing_summary here, (avoid window too long for data)
                #  Desirable to have some "processing_run" iterator supplied by KernelDataset.
                from aurora.pipelines.time_series_helpers import (  # TODO: consider storing clock-zero-truncated data
                    truncate_to_clock_zero,
                )

                tmp = tfk_dataset.df.copy(deep=True)
                group_by = [
                    "start",
                ]
                grouper = tmp.groupby(
                    group_by
                )  # these should have 1 or two rows per group
                for start, df in grouper:
                    end = df.end.unique()[0]  # nice to have this for info log
                    logger.debug("Access ch1 and ch2 ")
                    ch1_row = df[df.station == feature.station_1].iloc[0]
                    ch1_data = ch1_row.run_dataarray.to_dataset("channel")[
                        feature.channel_1
                    ]
                    ch1_data = truncate_to_clock_zero(
                        decimation_obj=dec_level_config, run_xrds=ch1_data
                    )
                    ch2_row = df[df.station == feature.station_2].iloc[0]
                    ch2_data = ch2_row.run_dataarray.to_dataset("channel")[
                        feature.channel_2
                    ]
                    ch2_data = truncate_to_clock_zero(
                        decimation_obj=dec_level_config, run_xrds=ch2_data
                    )
                    msg = f"Data for computing {feature.name} on {start} -- {end} ready"
                    logger.debug(msg)
                    # Compute the feature.
                    freqs, coherence_spectrogram = feature.compute(ch1_data, ch2_data)
                    # TODO: consider making get_time_axis() a method of the feature class
                    # Get Time Axis (See Development Note #4)
                    # now create a WindowedTimeSeries and get the time axis")
                    from aurora.time_series.windowing_scheme import (
                        window_scheme_from_decimation,
                    )

                    windowing_scheme = window_scheme_from_decimation(
                        decimation=dec_level_config
                    )
                    ch1_dataset = ch1_data.to_dataset()
                    windowed_obj = windowing_scheme.apply_sliding_window(
                        ch1_dataset, dt=1.0 / dec_level_config.decimation.sample_rate
                    )
                    # Now, take the time axis off the windowed obj, and bind it to coherence_spectrogram
                    # Check the logic in time_series_helpers.fft_xr_ds for how to do this, and note there is a model
                    # there for multivariate.
                    coherence_spectrogram_xr = xr.DataArray(
                        coherence_spectrogram,
                        dims=["time", "frequency"],
                        coords={"frequency": freqs, "time": windowed_obj.time.data},
                    )
                    feature_chunks.append(coherence_spectrogram_xr)
                feature_data = xr.concat(feature_chunks, "time")
                # should fill NaNs with 0s, otherwise thing break downstream.
                feature_data = feature_data.fillna(0)
                feature.data = feature_data  # bind feature data to feature instance (maybe temporal workaround)

                logger.info(f"Feature {feature.name} computed. Data has shape {feature_data.shape}")

    return


def read_features_from_mth5():
    raise NotImplementedError


def calculate_weights(
    dec_level_config: AuroraDecimationLevel, tfk_dataset: KernelDataset
):
    """
        Placeholder.

        Development Note #5.
            The calculation of weights from features is in general not a simple operation.
            Naive calculation by multiplying data with weight functions may be relatively simple, but even this is
            complicated by the selection of weight functions, and transition regions.  Moreover, when using hard
            thresholds (weights that can go to zero) it is possible to unintentionally "zero-out" the whole dataset.

            Such problems can be partly addressed by implementing strategies such as; if all weights are zero, keep
            the "best 20%" of the data, say.  However, when multiple features are being used, we then need to handle
            ranking of data that may fail to qualify based on one feature, but OK based on another. General handling
            for this may involve case studies of features and their relationship to data quality, and should be done
            with visualization capability.

            This analysis can be further complicated by the fact that some features are dynamic.  Consider
            for example the Mahalnobis Distance (MD).  This feature gives a reasonable way to down-weight data based on
            the empirical distributions, but, say that another feature is also employed (multiple coherence (MC) for
            example). The MD values of the data will change depending on whether the MC was applied before or
            after the MC trimming of the dataset.  In the MD case, a reasonable workaround would be to add an MD option
            in the regression, rather than in the preliminary weighting, since the robust regression repeatedly
            recomputes weights based on updated distributions.  Thus features can be divided into 'static' and
            'dynamic'.  Static features can be computed before regression, and used to pre-weight the data, whereas
            dynamic features are better worked into the regression loops.

            There may not be an ideal general solution, but it is likely that there will be a few "plays" that we
            can devise that will work to improve the data most of the time.

            For the reasons outlined above, the weight calculation will be in development for some time, and will
            likely require feature-visualization tools before we settle on a fixed set of strategies.  In the meantime,
            we can support

    Parameters
    ----------
    features
    chws

    Returns
    -------

    """

    # loop the channel weight specs
    for chws in dec_level_config.channel_weight_specs:
        msg = f"{chws}"
        logger.debug(msg)
        # TODO: Consider calculating all the weight kernels in advance, case switching on the combination style.
        if chws.combination_style == "multiplication":
            print(f"chws.combination_style {chws.combination_style}")

            weights = None
            # loop the feature weight specs
            for fws in chws.feature_weight_specs:
                if fws.weight_kernels is None:
                    msg = f"Feature weight spec {fws} has no weight kernels defined, skipping"
                    logger.warning(msg)
                    continue  # skip this feature weight spec
                msg = f"feature weight spec: {fws}"
                logger.debug(msg)
                feature = fws.feature
                msg = f"feature: {feature}"
                logger.debug(msg)
                # TODO: confirm that the feature object has its data
                #print("feature.data", feature.data, len(feature.data))

                # TODO: Now apply the fws weighting to the feature data
                #  Hopefully this is independent of the feature.
                weights = None
                for wk in fws.weight_kernels:
                    if weights is None:
                        weights = wk.evaluate(feature.data)
                    else:
                        weights *= wk.evaluate(feature.data)
                # chws.weights[fws.feature.name] = weights
            chws.weights = weights
            logger.info(f"Computed weights for {str(chws.output_channels)} using {str(chws.combination_style)} combination style.")

        else:
            msg = f"chws.combination_style {str(chws.combination_style)} not implemented"
            raise ValueError(msg)

    return



"""
Module to compare two transfer functions.

"""

import pathlib
from typing import Union

import numpy as np
from loguru import logger
from matplotlib import pyplot as plt
from mt_metadata.transfer_functions.core import TF
from scipy.interpolate import interp1d


class CompareTF:
    def __init__(
        self,
        tf_01: Union[str, pathlib.Path, TF],
        tf_02: Union[str, pathlib.Path, TF],
    ):
        """
        Class to compare two transfer functions.

        Parameters
        ----------
        tf_01
            First transfer function (file path or TF object)
        tf_02
            Second transfer function (file path or TF object)
        """
        self._comp_dict = {
            1: "$Z_{xx}$",
            2: "$Z_{xy}$",
            3: "$Z_{yx}$",
            4: "$Z_{yy}$",
        }

        self._compare_keys = [
            "impedance_amplitude_close",
            "impedance_phase_close",
            "impedance_error_close",
            "impedance_ratio",
            "impedance_std",
            "impedance_correlation",
            "tipper_amplitude_close",
            "tipper_phase_close",
            "tipper_error_close",
            "tipper_ratio",
            "tipper_correlation",
            "tipper_std",
        ]

        self._impedance_keys = [
            ckey for ckey in self._compare_keys if "impedance" in ckey
        ]
        self._tipper_keys = [ckey for ckey in self._compare_keys if "tipper" in ckey]

        if isinstance(tf_01, (str, pathlib.Path)):
            self.tf_01 = TF()
            self.tf_01.read(tf_01)
        elif isinstance(tf_01, TF):
            self.tf_01 = tf_01
        else:
            raise TypeError("tf_01 must be a file path or TF object")

        if isinstance(tf_02, (str, pathlib.Path)):
            self.tf_02 = TF()
            self.tf_02.read(tf_02)
        elif isinstance(tf_02, TF):
            self.tf_02 = tf_02
        else:
            raise TypeError("tf_02 must be a file path or TF object")

    def plot_two_transfer_functions(
        self,
        label_01="emtf",
        label_02="aurora",
        save_plot_path=None,
        rho_xx_ylims=None,
        rho_xy_ylims=None,
        rho_yx_ylims=None,
        rho_yy_ylims=None,
        phi_ylims=None,
    ):
        """
        Plots two transfer functions for comparison.

        Parameters
        ----------
        label_01
            Label for the first transfer function
        label_02
            Label for the second transfer function
        save_plot_path
            Path to save the plot (optional)

        Returns
        -------

        """
        xy = "xy"
        fig = plt.figure(figsize=(12, 6))

        for ii in range(2):
            for jj in range(2):
                component = f"{xy[ii]}{xy[jj]}"
                plot_num_res = 1 + ii * 2 + jj
                plot_num_phase = 5 + ii * 2 + jj
                ax = fig.add_subplot(2, 4, plot_num_res)
                ax.loglog(
                    self.tf_01.period,
                    0.2
                    * self.tf_01.period
                    * np.abs(self.tf_01.impedance.data[:, ii, jj]) ** 2,
                    label=label_01,
                    marker="s",
                    markersize=7,
                    color="k",
                )
                ax.loglog(
                    self.tf_02.period,
                    0.2
                    * self.tf_02.period
                    * np.abs(self.tf_02.impedance.data[:, ii, jj]) ** 2,
                    label=label_02,
                    marker="o",
                    markersize=4,
                    color="r",
                )
                ax.set_title(self._comp_dict[plot_num_res])
                # ax.set_xlabel("Period (s)")
                if plot_num_res == 1:
                    ax.set_ylabel("Apparent Resistivity ($\\Omega \\cdot m$)")
                    ax.legend()
                ax.grid(True, which="both", ls="--", lw=0.5, color="gray")
                if component == "xx" and rho_xx_ylims is not None:
                    ax.set_ylim(rho_xx_ylims)
                if component == "xy" and rho_xy_ylims is not None:
                    ax.set_ylim(rho_xy_ylims)
                if component == "yx" and rho_yx_ylims is not None:
                    ax.set_ylim(rho_yx_ylims)
                if component == "yy" and rho_yy_ylims is not None:
                    ax.set_ylim(rho_yy_ylims)

                ax2 = fig.add_subplot(2, 4, plot_num_phase)
                ax2.semilogx(
                    self.tf_01.period,
                    np.degrees(np.angle(self.tf_01.impedance.data[:, ii, jj])),
                    label=label_01,
                    marker="s",
                    markersize=7,
                    color="k",
                )
                ax2.semilogx(
                    self.tf_02.period,
                    np.degrees(np.angle(self.tf_02.impedance.data[:, ii, jj])),
                    label=label_02,
                    marker="o",
                    markersize=4,
                    color="r",
                )
                ax2.set_xlabel("Period (s)")
                if plot_num_phase == 5:
                    ax2.set_ylabel("Phase (degrees)")
                    ax2.legend()

                ax2.grid(True, which="both", ls="--", lw=0.5, color="gray")

        fig.tight_layout()
        plt.show()

        if save_plot_path is not None:
            fig.savefig(save_plot_path, dpi=300)
            logger.info(f"Saved comparison plot to {save_plot_path}")
            plt.close(fig)

    def _interpolate_complex_array(
        self,
        source_periods: np.ndarray,
        source_array: np.ndarray,
        target_periods: np.ndarray,
    ) -> np.ndarray:
        """Interpolate complex array onto target periods."""
        interp_array = np.zeros(
            (len(target_periods),) + source_array.shape[1:], dtype=complex
        )

        for i in range(source_array.shape[1]):
            for j in range(source_array.shape[2]):
                real_interp = interp1d(
                    source_periods,
                    source_array[:, i, j].real,
                    kind="linear",
                    bounds_error=False,
                    fill_value="extrapolate",
                )
                imag_interp = interp1d(
                    source_periods,
                    source_array[:, i, j].imag,
                    kind="linear",
                    bounds_error=False,
                    fill_value="extrapolate",
                )
                interp_array[:, i, j] = real_interp(target_periods) + 1j * imag_interp(
                    target_periods
                )

        return interp_array

    def interpolate_tf_to_common_periods(self):
        """
        Interpolate two transfer functions onto common period range.

        Uses the overlapping period range and creates a common grid for comparison.

        Parameters
        ----------
        tf1 : TF
            First transfer function
        tf2 : TF
            Second transfer function

        Returns
        -------
        periods_common : ndarray
            Common period array
        z1_interp : ndarray
            Interpolated impedance from tf1, shape (n_periods, 2, 2)
        z2_interp : ndarray
            Interpolated impedance from tf2, shape (n_periods, 2, 2)
        z1_err_interp : ndarray
            Interpolated impedance errors from tf1
        z2_err_interp : ndarray
            Interpolated impedance errors from tf2
        """
        # Get period arrays
        p1 = self.tf_01.period
        p2 = self.tf_02.period

        # Find overlapping range
        p_min = max(p1.min(), p2.min())
        p_max = min(p1.max(), p2.max())

        # Create common period grid (logarithmic spacing)
        n_periods = min(len(p1), len(p2))
        periods_common = np.logspace(np.log10(p_min), np.log10(p_max), n_periods)

        if self.tf_01.has_impedance() and self.tf_02.has_impedance():
            # Interpolate tf1 impedance (log-log for real and imag separately)
            z1_interp = self._interpolate_complex_array(
                p1, self.tf_01.impedance, periods_common
            )
            z1_err_interp = self._interpolate_complex_array(
                p1, self.tf_01.impedance_error, periods_common
            )

            z2_interp = self._interpolate_complex_array(
                p2, self.tf_02.impedance, periods_common
            )
            z2_err_interp = self._interpolate_complex_array(
                p2, self.tf_02.impedance_error, periods_common
            )
        else:
            z1_interp = None
            z2_interp = None
            z1_err_interp = None
            z2_err_interp = None

        if self.tf_01.has_tipper() and self.tf_02.has_tipper():
            t1_interp = self._interpolate_complex_array(
                p1, self.tf_01.tipper, periods_common
            )
            t2_interp = self._interpolate_complex_array(
                p2, self.tf_02.tipper, periods_common
            )
            t1_err_interp = self._interpolate_complex_array(
                p1, self.tf_01.tipper_error, periods_common
            )
            t2_err_interp = self._interpolate_complex_array(
                p2, self.tf_02.tipper_error, periods_common
            )
        else:
            t1_interp = None
            t2_interp = None
            t1_err_interp = None
            t2_err_interp = None

        return (
            periods_common,
            z1_interp,
            z2_interp,
            z1_err_interp,
            z2_err_interp,
            t1_interp,
            t2_interp,
            t1_err_interp,
            t2_err_interp,
        )

    def compare_transfer_functions(
        self,
        rtol: float = 1,
        atol: float = 1,
        atol_phase: float = 4.0,
    ) -> dict:
        """
        Compare transfer functions between two transfer_functions objects.

        Compares transfer_functions, sigma_e, and sigma_s arrays. If periods
        don't match, interpolates one onto the other.

        Parameters
        ----------
        rtol: float
            Relative tolerance for np.allclose, defaults to 1e-2
        atol: float
            Absolute tolerance for np.allclose, defaults to 1e-2
        atol_phase: float
            Absolute tolerance for phase comparison, defaults to 4.0 degrees

        Returns
        -------
        comparison: dict
            Dictionary containing:
            - "periods_match": bool, whether periods are identical
            - "transfer_functions_close": bool
            - "sigma_e_close": bool
            - "sigma_s_close": bool
            - "max_tf_diff": float, max absolute difference in transfer functions
            - "max_sigma_e_diff": float
            - "max_sigma_s_diff": float
            - "periods_used": np.ndarray of periods used for comparison
        """

        (
            periods_common,
            z1,
            z2,
            z1_err,
            z2_err,
            t1,
            t2,
            t1_err,
            t2_err,
        ) = self.interpolate_tf_to_common_periods()

        result = dict([(key, None) for key in self._compare_keys])

        result["periods_used"] = periods_common

        # Compare arrays
        if z1 is not None and z2 is not None:
            for ckey in self._impedance_keys:
                result[ckey] = {}

            for ii in range(2):
                for jj in range(2):
                    ratio = np.median(np.abs(z1[:, ii, jj]) / np.abs(z2[:, ii, jj]))
                    key = f"Z_{ii}{jj}"
                    result["impedance_ratio"][key] = ratio
                    result["impedance_correlation"][key] = np.corrcoef(
                        np.abs(z1[:, ii, jj]), np.abs(z2[:, ii, jj])
                    ).min()
                    result["impedance_std"][key] = np.std(
                        np.abs(z1[:, ii, jj] - z2[:, ii, jj])
                    )
                    result["impedance_amplitude_close"] = np.allclose(
                        np.abs(z1[:, ii, jj]),
                        np.abs(z2[:, ii, jj]),
                        rtol=rtol,
                        atol=atol,
                    )

                    result["impedance_phase_close"] = np.allclose(
                        np.angle(z1[:, ii, jj]),
                        np.angle(z2[:, ii, jj]),
                        rtol=rtol,
                        atol=atol_phase,
                    )

                    result["impedance_error_close"] = np.allclose(
                        np.abs(z1_err[:, ii, jj]),
                        np.abs(z2_err[:, ii, jj]),
                        rtol=rtol,
                        atol=atol,
                    )

        if t1 is not None and t2 is not None:
            for ckey in self._tipper_keys:
                result[ckey] = {}

            for ii in range(1):
                for jj in range(2):
                    ratio = np.median(np.abs(t1[:, ii, jj]) / np.abs(t2[:, ii, jj]))
                    key = f"T_{ii}{jj}"
                    result["tipper_ratio"][key] = ratio
                    result["tipper_correlation"][key] = np.corrcoef(
                        np.abs(t1[:, ii, jj]), np.abs(t2[:, ii, jj])
                    ).min()
                    result["tipper_std"][key] = np.std(
                        np.abs(t1[:, ii, jj] - t2[:, ii, jj])
                    )
                    result["tipper_amplitude_close"] = np.allclose(
                        np.abs(t1[:, ii, jj]),
                        np.abs(t2[:, ii, jj]),
                        rtol=rtol,
                        atol=atol,
                    )

                    result["tipper_phase_close"] = np.allclose(
                        np.angle(t1[:, ii, jj]),
                        np.angle(t2[:, ii, jj]),
                        rtol=rtol,
                        atol=atol,
                    )

                    result["tipper_error_close"] = np.allclose(
                        np.abs(t1_err[:, ii, jj]),
                        np.abs(t2_err[:, ii, jj]),
                        rtol=rtol,
                        atol=atol,
                    )

        return result



"""
Integrated test of the functionality of feature weights.

1. This test uses degraded synthetic data to test the feature weighting.
Noise is added to some fraction (50-75%) of the data.

Then regular (single station) processing is called on the data and
feature weighting processing is called on the data.

---
Feature weights are specified using the mt_metadata.features.weights module.
This test demonstrates how feature-based channel weighting (e.g., striding_window_coherence)
can be injected into Aurora's processing pipeline. In the future, these features will be
used to enable more robust, data-driven weighting strategies for transfer function estimation,
including integration of new features from mt_metadata and more flexible weighting schemes.

See also: mt_metadata.features.weights.channel_weight_spec and test_feature_weighting.py for
examples of how to define, load, and use feature weights in Aurora workflows.
"""

import json
import pathlib
from typing import Optional

import numpy as np
from loguru import logger
from mt_metadata.features.weights.channel_weight_spec import ChannelWeightSpec
from mt_metadata.transfer_functions import TF
from mth5.mth5 import MTH5
from mth5.processing import KernelDataset, RunSummary
from mth5.timeseries import RunTS

from aurora.config.metadata import Processing
from aurora.config.metadata.processing import _processing_obj_from_json_file
from aurora.general_helper_functions import (
    MT_METADATA_FEATURES_TEST_HELPERS_PATH,
    PROCESSING_TEMPLATES_PATH,
    TEST_PATH,
)
from aurora.pipelines.process_mth5 import process_mth5


def create_synthetic_mth5_with_noise(
    source_file: pathlib.Path,
    target_file: Optional[pathlib.Path] = None,
    noise_channels=("ex", "hy"),
    frac=0.5,
    noise_level=1000.0,
    seed=None,
):
    """
    Copy a synthetic MTH5, injecting noise into specified channels for a fraction of the data.
    """
    if target_file is None:
        target_file = TEST_PATH.joinpath("synthetic", "test1_noisy.h5")
        if target_file.exists():
            target_file.unlink()
    if seed is None:
        seed = 42  # Default seed for reproducibility

    rng = np.random.default_rng(seed)
    m_source = MTH5(source_file)
    m_source.open_mth5(mode="r")
    m_target = MTH5(target_file, file_version=m_source.file_version)
    m_target.open_mth5(mode="w")

    for station_id in m_source.station_list:
        station = m_source.get_station(station_id)
        if station_id not in m_target.station_list:
            m_target.add_station(station_id, station_metadata=station.metadata)
        for run_id in station.run_summary["id"].unique():
            run = station.get_run(run_id)
            ch_list = []
            for ch in run.channel_summary.component.to_list():
                ch_obj = run.get_channel(ch)
                ch_ts = ch_obj.to_channel_ts()
                data = ch_ts.data_array.data.copy()
                n = len(data)
                if ch in noise_channels:
                    noisy_idx = slice(0, int(frac * n))
                    noise = rng.normal(0, noise_level, size=data[noisy_idx].shape)
                    noise = noise.astype(
                        data.dtype
                    )  # Ensure noise is the same dtype as data
                    data[noisy_idx] += noise
                ch_ts.data_array.data = data
                ch_list.append(ch_ts)
            runts = RunTS(array_list=ch_list, run_metadata=run.metadata)
            runts.run_metadata.id = run_id
            target_station = m_target.get_station(station_id)
            target_station.add_run(run_id).from_runts(runts)
    m_source.close_mth5()
    m_target.close_mth5()
    return target_file


def _load_example_channel_weight_specs(
    keep_only=[
        "striding_window_coherence",
    ]
) -> list:
    """
    Loads example channel weight specifications from a JSON file.

    Modifies it for this test so that the feature_weight_specs are only striding_window_coherence.

    Parameters
    ----------
    keep_only: list
        List of feature names to keep in the feature_weight_specs.
        Default is ["striding_window_coherence"].
    Returns
    -------
    output: list
        List of ChannelWeightSpec objects with modified feature_weight_specs.
    """
    feature_weight_json = MT_METADATA_FEATURES_TEST_HELPERS_PATH.joinpath(
        "channel_weight_specs_example.json"
    )
    assert (
        feature_weight_json.exists()
    ), f"Could not find feature weighting block json at {feature_weight_json}"

    with open(feature_weight_json, "r") as f:
        data = json.load(f)

    output = []
    channel_weight_specs = data.get("channel_weight_specs", data)
    for cws_dict in channel_weight_specs:
        # Unwrap the nested structure
        cws_data = cws_dict.get("channel_weight_spec", cws_dict)

        # Process feature_weight_specs to unwrap nested dicts
        if "feature_weight_specs" in cws_data:
            fws_list = []
            for fws_item in cws_data["feature_weight_specs"]:
                fws_data = fws_item.get("feature_weight_spec", fws_item)
                fws_list.append(fws_data)
            cws_data["feature_weight_specs"] = fws_list

        # Construct directly from dict to ensure proper deserialization
        cws = ChannelWeightSpec(**cws_data)

        # Modify the feature_weight_specs to only include striding_window_coherence
        if keep_only:
            cws.feature_weight_specs = [
                fws for fws in cws.feature_weight_specs if fws.feature.name in keep_only
            ]
            # get rid of Remote reference channels (work in progress)
            cws.feature_weight_specs = [
                fws for fws in cws.feature_weight_specs if fws.feature.channel_2 != "rx"
            ]
            cws.feature_weight_specs = [
                fws for fws in cws.feature_weight_specs if fws.feature.channel_2 != "ry"
            ]

        # Ensure that the feature_weight_specs is not empty
        if not cws.feature_weight_specs:
            msg = "No valid feature_weight_specs found in channel weight spec."
            logger.error(msg)
        else:
            output.append(cws)

    return output


def load_processing_objects() -> dict:
    """
    Loads the 'default' and 'with_weights' processing objects.

    'default' is loaded from the processing configuration template.
    'with_weights' is loaded from the same template but with channel weight specs
    set to only include 'striding_window_coherence'.

    Returns
    -------
    dict
        Dictionary with keys 'default' and 'with_weights' mapping to Processing objects.
    """
    processing_params_json = PROCESSING_TEMPLATES_PATH.joinpath(
        "processing_configuration_template.json"
    )

    processing_objects = {}
    processing_objects["default"] = _processing_obj_from_json_file(
        processing_params_json
    )

    cws_list = _load_example_channel_weight_specs(
        keep_only=[
            "striding_window_coherence",
        ]
    )
    processing_objects["with_weights"] = _processing_obj_from_json_file(
        processing_params_json
    )
    processing_objects["with_weights"].decimations[0].channel_weight_specs = cws_list

    return processing_objects


def process_mth5_with_config(
    mth5_path: pathlib.Path, processing_obj: Processing, z_file="test1.zss"
) -> TF:
    """
    Executes aurora processing on mth5_path, and returns mt_metadata TF object.
    """
    run_summary = RunSummary()
    run_summary.from_mth5s(list((mth5_path,)))

    kernel_dataset = KernelDataset()
    kernel_dataset.from_run_summary(run_summary, "test1")
    config = processing_obj
    config.stations.remote = []  # TODO: allow this to be False
    for dec in config.decimations:
        dec.estimator.engine = "RME"
        dec.reference_channels = []

    tf_cls = process_mth5(
        config,
        kernel_dataset,
        units="MT",
        z_file_path=z_file,
        show_plot=False,
    )
    return tf_cls


def print_apparent_resistivity(tf, label="TF"):
    """
    Print apparent resistivity and phase for each period/frequency in the TF object.
    Returns the mean apparent resistivity (averaged over all frequencies and both Zxy/Zyx).
    """
    if not hasattr(tf, "impedance"):
        print(f"{label}: TF object missing impedance attribute.")
        return np.nan
    z = tf.impedance
    print(
        f"{label} impedance shape: {getattr(z, 'shape', None)}, dims: {getattr(z, 'dims', None)}"
    )

    # Get period and convert to frequency
    if hasattr(tf, "period"):
        period = np.array(tf.period)
        freq = 1.0 / period
    elif hasattr(tf, "frequency"):
        freq = np.array(tf.frequency)
    else:
        print(f"{label}: TF object missing period/frequency attribute.")
        return np.nan

    n_periods = z.shape[0]
    n_out = z.shape[1]
    n_in = z.shape[2]
    print(
        f"{label} n_periods={n_periods}, n_out={n_out}, n_in={n_in}, len(freq)={len(freq)}"
    )

    rho_vals = []
    for i in range(min(n_periods, len(freq))):
        f = freq[i]
        for comp, out_idx, in_idx in [("Zxy", 0, 1), ("Zyx", 1, 0)]:
            if out_idx < n_out and in_idx < n_in:
                zval = z[i, out_idx, in_idx]
                rho = (np.abs(zval) ** 2) / (2 * np.pi * f)
                phase = np.angle(zval, deg=True)
                print(
                    f"{label} f={f:.4g} Hz {comp}: rho={rho:.3g} ohm-m, phase={phase:.2f} deg"
                )
                rho_vals.append(rho)
            else:
                print(
                    f"{label} index out of bounds: out_idx={out_idx}, in_idx={in_idx}"
                )
    mean_rho = np.nanmean(rho_vals) if rho_vals else np.nan
    print(
        f"{label} MEAN apparent resistivity (all freqs, Zxy/Zyx): {mean_rho:.3g} ohm-m"
    )
    return mean_rho


def test_feature_weighting(synthetic_test_paths, worker_safe_test1_h5):
    """Test that feature weighting affects TF processing results."""
    SYNTHETIC_FOLDER = synthetic_test_paths.aurora_results_path.parent

    # Create a synthetic mth5 file for testing
    mth5_path = create_synthetic_mth5_with_noise(source_file=worker_safe_test1_h5)

    processing_objects = load_processing_objects()
    z_path1 = SYNTHETIC_FOLDER.joinpath("test1_default.zss")
    z_path2 = SYNTHETIC_FOLDER.joinpath("test1_weights.zss")
    process_mth5_with_config(mth5_path, processing_objects["default"], z_file=z_path1)
    process_mth5_with_config(
        mth5_path, processing_objects["with_weights"], z_file=z_path2
    )

    tf1 = TF(fn=z_path1)
    tf2 = TF(fn=z_path2)
    tf1.read(**{"rotate_to_measurement_coordinates": False})
    tf2.read(**{"rotate_to_measurement_coordinates": False})

    assert (
        tf1.impedance.data != tf2.impedance.data
    ).any(), "TF1 and TF2 should have different impedance values after processing with weights."

    print("TF1 Apparent Resistivity and Phase:")
    mean_rho1 = print_apparent_resistivity(tf1, label="TF1")
    print("TF2 Apparent Resistivity and Phase:")
    mean_rho2 = print_apparent_resistivity(tf2, label="TF2")
    print(
        f"\nSUMMARY: Mean apparent resistivity TF1: {mean_rho1:.3g} ohm-m, TF2: {mean_rho2:.3g} ohm-m"
    )



"""

    This module contains the regression M-estimator for remote reference regression.

Development Notes:
- follows Gary's TRME.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes


"""
import numpy as np
import xarray as xr
from loguru import logger

from aurora.transfer_function.regression.m_estimator import MEstimator


class RME_RR(MEstimator):
    def __init__(self, **kwargs):
        """
        Constructor.

        Robust remote reference estimator.  Z is the reference station data,
        and is the same size as X (see regression.base.RegressionEstimator(
        """
        super(RME_RR, self).__init__(**kwargs)
        self._Z = kwargs.get("Z", None)
        self.Z = self._Z.to_array().data.T
        self.qr_input = "Z"
        self._QHX = None
        self.check_for_nan()
        self.check_for_enough_data_for_rr_estimate()
        self.check_reference_data_shape()

    def check_for_nan(self):
        """Raises exception if NaN in any of the regression data"""
        cond1 = np.isnan(self.X).any()
        cond2 = np.isnan(self.Y).any()
        cond3 = np.isnan(self.Z).any()
        nans_present = cond1 or cond2 or cond3
        if nans_present:
            logger.error("Missing data not allowed for RME_RR class")
            raise Exception

    def check_for_enough_data_for_rr_estimate(self) -> None:
        """Raises exception if not enough data for remote reference estimate"""
        if self.is_underdetermined:
            error_msg = "not enough data for RR estimate:"
            error_msg = f"{error_msg} n_channels_in = {self.n_channels_in}"
            error_msg = f"{error_msg} N_data = {self.n_data}"
            logger.error(f"{error_msg}")
            raise Exception

    def check_reference_data_shape(self) -> None:
        """Raises exception if data shape check fails."""
        if self.Z.shape != self.X.shape:
            logger.error(
                "sizes of local and remote do not agree in RR estimation routine"
            )
            raise Exception

    def update_y_hat(self) -> None:
        """Updates the predicted data."""
        self._Y_hat = self.X @ self.b

    def update_b(self) -> None:
        r"""
        Updates the tf estimate data.

        matlab code was: b = QTX\QTY
        """
        self.b = np.linalg.solve(self.QHX, self.QHYc)
        return

    @property
    def QHX(self) -> np.ndarray:
        """
        Returns the matrix QH @ X
        """
        if self._QHX is None:
            self._QHX = self.QH @ self.X
        return self._QHX

    def update_residual_variance(self, correction_factor=1) -> np.ndarray:
        """updates residual variance from most recent cleaned and predicted data"""
        self._residual_variance = self.residual_variance_method1()
        self._residual_variance *= correction_factor
        return self._residual_variance

    def compute_inverse_signal_covariance(self) -> xr.DataArray:
        r"""
        Computes the inverse signal covariance matrix of the input channels.

        Development Notes:
        Original Matlab code was basically this one-liner:
        Cov_SS = (self.ZH @self.X) \ (self.XH @ self.X) / (self.XH @self.Z)
        I broke the above line into B/A where
        B = (self.ZH @self.X) \ (self.XH @ self.X), and A = (self.XH @self.Z)
        Then follow matlab cookbook, B/A for matrices = (A'\B')
        """
        ZH = self.Z.conj().T
        XH = self.X.conj().T
        B = np.linalg.solve(ZH @ self.X, XH @ self.X)
        A = XH @ self.Z
        cov_ss_inv = np.linalg.solve(A.conj().T, B.conj().T).conj().T

        self.cov_ss_inv = xr.DataArray(
            cov_ss_inv,
            dims=["input_channel_1", "input_channel_2"],
            coords={
                "input_channel_1": list(self._X.data_vars),
                "input_channel_2": list(self._X.data_vars),
            },
        )
        return



"""
This module contains the TrasnferFunctionKernel class which is the main object that
links the KernelDataset to Processing configuration.

"""

import pathlib
from typing import List, Union

import numpy as np
import pandas as pd
import psutil
from loguru import logger
from mt_metadata.processing.aurora import DecimationLevel as AuroraDecimationLevel
from mt_metadata.transfer_functions.core import TF
from mth5.processing.kernel_dataset import KernelDataset
from mth5.utils.exceptions import MTH5Error
from mth5.utils.helpers import path_or_mth5_object

from aurora import __version__ as aurora_version
from aurora.config.metadata.processing import Processing
from aurora.pipelines.helpers import initialize_config
from aurora.pipelines.time_series_helpers import prototype_decimate
from aurora.time_series.windowing_scheme import WindowingScheme
from aurora.transfer_function import TransferFunctionCollection


class TransferFunctionKernel(object):
    def __init__(
        self,
        dataset: KernelDataset,
        config: Union[Processing, str, pathlib.Path],
    ):
        """
        Constructor.

        Parameters
        ----------
        dataset: mth5.processing.kernel_dataset.KernelDataset
         Specification of the dataset that will be processed to yield a transfer function.
        config: aurora.config.metadata.processing.Processing
         Specification of the processing parameters to be applied to the dataset.
        """
        processing_config = initialize_config(config)
        self._config = processing_config
        self._dataset = dataset
        self._mth5_objs = None
        self._memory_warning = False

    @property
    def dataset(self) -> KernelDataset:
        """returns the KernelDataset object"""
        return self._dataset

    @property
    def kernel_dataset(self) -> KernelDataset:
        """returns the KernelDataset object"""
        return self._dataset

    @property
    def dataset_df(self) -> pd.DataFrame:
        """returns the KernelDataset dataframe"""
        return self._dataset.df

    @property
    def processing_config(self) -> Processing:
        """Returns the processing config object"""
        return self._config

    @property
    def config(self) -> Processing:
        """Returns the processing config object"""
        return self._config

    @property
    def processing_summary(self) -> pd.DataFrame:
        """Returns the processing summary object -- creates if it doesn't yet exist."""
        if self._processing_summary is None:
            self.update_processing_summary()
        return self._processing_summary

    @property
    def mth5_objs(self) -> dict:
        """Returns self._mth5_objs"""
        if self._mth5_objs is None:
            self.initialize_mth5s()
        return self._mth5_objs

    def initialize_mth5s(self):
        """
        returns a dict of open mth5 objects, keyed by station_id

        A future version of this for multiple station processing may need nested dict with [survey_id][station]

        Returns
        -------
        mth5_objs : dict
            Keyed by stations.
            local station id : mth5.mth5.MTH5
            remote station id: mth5.mth5.MTH5
        """
        mode = self.get_mth5_file_open_mode()
        self._mth5_objs = self.kernel_dataset.initialize_mth5s(mode)

    def update_dataset_df(self, i_dec_level: int) -> None:
        """
        This function has two different modes.  The first mode initializes values in the
        array, and could be placed into TFKDataset.initialize_time_series_data()
        The second mode, decimates. The function is kept in pipelines because it calls
        time series operations.

        Notes:
        1. When assigning xarrays to dataframe cells, df dislikes xr.Dataset,
        so we convert to DataArray before assignment


        Parameters
        ----------
        i_dec_level: int
            decimation level id, indexed from zero

        Returns
        -------
        dataset_df: pd.DataFrame
            Same df that was input to the function but now has columns:


        """
        if i_dec_level == 0:
            # ANY MERGING OF RUNS IN TIME DOMAIN WOULD GO HERE

            # Assign additional columns to dataset_df, populate with mth5_objs and xr_ts
            self.dataset.initialize_dataframe_for_processing()

            # APPLY TIMING CORRECTIONS HERE
        else:
            logger.info(f"DECIMATION LEVEL {i_dec_level}")

            for i, row in self.dataset_df.iterrows():
                if not self.is_valid_dataset(row, i_dec_level):
                    continue
                if row.fc:
                    row_ssr_str = (
                        f"survey: {row.survey}, station: {row.station}, run: {row.run}"
                    )
                    msg = f"FC already exists for {row_ssr_str} -- skipping decimation"
                    logger.info(msg)
                    continue
                run_xrds = row["run_dataarray"].to_dataset("channel")
                decimation = self.config.decimations[i_dec_level].decimation
                decimated_xrds = prototype_decimate(decimation, run_xrds)
                self.dataset_df.at[i, "run_dataarray"] = decimated_xrds.to_array(
                    "channel"
                )  # See Note 1 above

        logger.info(
            f"Dataset Dataframe Updated for decimation level {i_dec_level} Successfully"
        )
        return

    def apply_clock_zero(self, dec_level_config: AuroraDecimationLevel):
        """
        get clock-zero from data if needed

        Parameters
        ----------
        dec_level_config: AuroraDecimationLevel
            metadata about the decimation level processing.
        Returns
        -------
        dec_level_config: AuroraDecimationLevel
            The modified DecimationLevel with clock-zero information set.
        """
        if dec_level_config.stft.window.clock_zero_type == "data start":
            dec_level_config.stft.window.clock_zero = str(self.dataset_df.start.min())
        return dec_level_config

    @property
    def all_fcs_already_exist(self) -> bool:
        """Return true of all FCs needed to process data already exist in the mth5s"""
        if self.kernel_dataset.df["fc"].isna().any():
            self.check_if_fcs_already_exist()

        # these should all be booleans now
        assert not self.kernel_dataset.df["fc"].isna().any()

        return self.kernel_dataset.df.fc.all()

    def check_if_fcs_already_exist(self):
        """
        Fills out the "fc" column of dataset dataframe with True/False.

        If all FC Levels for a given station-run are already built, mark True otherwise False.

        Iterates over the processing summary_df, grouping by unique "Survey-Station-Run"s.
        (Could also iterate over kernel_dataset.dataframe, to get the groupby).

        Note 1:  Because decimation is a cascading operation, we avoid the case where some (valid) decimation
        levels exist in the mth5 FC archive and others do not.  The maximum granularity is the
        "station-run" level. For a given run, either all relevant FCs are in the h5 or we treat as if none
        of them are.  To support variations at the decimation-level, an appropriate way to address would
        be to store decimated time series in the archive as well (they would simply be runs with different sample
        rates, and some extra filters).

        Note #2: run_sub_df may have multiple rows, even though the run id is unique.
        This could happen for example when you have a long run at the local station, but multiple (say two) shorter runs
        at the reference station.  In that case, the processing summary will have a separate row for the
        intersection of the long run with each of the remote runs. We ignore this for now, selecting only the first
        element of the run_sub_df, under the assumption that FCs have been created for the entire run,
        or not at all.  This assumption can be relaxed in future by using the time_period attribute of the FC layer.
        For now, we proceed with the all-or-none logic.  That is, if a ['survey', 'station', 'run',] has FCs,
        assume that the FCs are present for the entire run. We assign the "fc" column of dataset_df to have the same
        boolean value for all rows of same  ['survey', 'station', 'run'] .


        Returns: None
            Modifies self.dataset_df inplace, assigning bools to the "fc" column

        """
        groupby = [
            "survey",
            "station",
            "run",
        ]
        grouper = self.processing_summary.groupby(groupby)

        for (survey_id, station_id, run_id), df in grouper:
            cond1 = self.dataset_df.survey == survey_id
            cond2 = self.dataset_df.station == station_id
            cond3 = self.dataset_df.run == run_id
            run_sub_df = self.dataset_df[cond1 & cond2 & cond3]

            if len(run_sub_df) > 1:
                # See Note #2
                msg = "Not all runs will process as a continuous chunk "
                msg += "-- in future may need to loop over runlets to check for FCs"
                logger.warning(msg)

            dataset_df_indices = np.r_[run_sub_df.index]
            remote = run_sub_df.remote.iloc[0]
            mth5_path = run_sub_df.mth5_path.iloc[0]
            fcs_present = mth5_has_fcs(
                mth5_path,
                survey_id,
                station_id,
                run_id,
                remote,
                self.processing_config,
                mode="r",
            )
            self.dataset_df.loc[dataset_df_indices, "fc"] = fcs_present

        if self.dataset_df["fc"].any():
            if self.dataset_df["fc"].all():
                msg = "All fc_levels already exist"
                msg += "Skip time series processing is OK"
            else:
                msg = f"Some, but not all fc_levels already exist = {self.dataset_df['fc']}"
        else:
            msg = "FC levels not present"
        logger.info(msg)

        return

    def show_processing_summary(
        self,
        omit_columns=(
            "mth5_path",
            "channel_scale_factors",
            "start",
            "end",
            "input_channels",
            "output_channels",
            "num_samples_overlap",
            "num_samples_advance",
            "run_dataarray",
        ),
    ):
        """
        Prints the processing summary table via logger.

        Parameters
        ----------
        omit_columns: tuple
            List of columns to omit when showing channel summary (used to keep table small).

        """
        columns_to_show = self.processing_summary.columns
        columns_to_show = [x for x in columns_to_show if x not in omit_columns]
        logger.info("Processing Summary Dataframe:")
        logger.info(f"\n{self.processing_summary[columns_to_show].to_string()}")

    def update_processing_summary(self):
        """
        Creates or updates the processing summary table based on processing parameters
        and kernel dataset.
        - Melt the decimation levels over the run summary.
        - Add columns to estimate the number of FFT windows for each row

        Returns
        -------
        processing_summary: pd.DataFrame
            One row per each run-deciamtion pair
        """
        processing_summary = _make_processing_summary(
            self.kernel_dataset.df,
            self.config.decimation_info,
            self.config.window_scheme(as_type="df"),
        )
        if processing_summary is not None:
            self._processing_summary = processing_summary
        else:
            msg = "Failed to update processing_summary"
            raise ValueError(msg)

    def validate_decimation_scheme_and_dataset_compatability(
        self, min_num_stft_windows=1
    ):
        """
        Checks that the decimation_scheme and dataset are compatable.
        Marks as invalid any rows that will fail to process based incompatibility.


        Refers to issue #182 (and #103, and possibly #196 and #233).
        Determine if there exist (one or more) runs that will yield decimated datasets
        that have too few samples to be passed to the STFT algorithm.

        Strategy for handling this:
        Mark as invlaid any rows of the processing summary that do not yield long
        enough time series to window.  This way all other rows, with decimations up to
        the invalid cases will still process.


        WCGW: If the runs are "chunked" we could encounter a situation where the
        chunk fails to yield a deep decimation level, yet the level could be produced
        if the chunk size were larger.
        In general, handling this seems a bit complicated.  We will ignore this for
        now and assume that the user will select a chunk size that is appropriate to
        the decimation scheme, i.e. use large chunks for deep decimations.

        A general solution: return a log that tells the user about each run and
        decimation level ... how many STFT-windows it yielded at each decimation level.
        This conjures the notion of (run, decimation_level) pairs
        -------

        """
        if min_num_stft_windows is None:
            min_stft_window_info = {
                x.decimation.level: x.stft.min_num_stft_windows
                for x in self.processing_config.decimations
            }
            min_stft_window_list = [
                min_stft_window_info[x] for x in self.processing_summary.dec_level
            ]
            min_num_stft_windows = pd.Series(min_stft_window_list)

        self.processing_summary["valid"] = (
            self.processing_summary.num_stft_windows >= min_num_stft_windows
        )

    def validate_processing(self):
        """
        Do some Validation checks. WIP.

        Things that are validated:
        1. The default estimation engine from the json file is "RME_RR", which is fine (
        we expect to in general to do more RR processing than SS) but if there is only
        one station (no remote)then the RME_RR should be replaced by default with "RME".
        - also if there is only one station, set reference channels to []

        2. make sure local station id is defined (correctly from kernel dataset)
        """

        # Make sure a RR method is not being called for a SS config
        if not self.config.stations.remote:
            self.config.drop_reference_channels()
            for decimation in self.config.decimations:
                if decimation.estimator.engine == "RME_RR":
                    logger.info("No RR station specified, switching RME_RR to RME")
                    decimation.estimator.engine = "RME"

        # Make sure that a local station is defined
        if not self.config.stations.local.id:
            logger.warning("WARNING: Local station not specified")
            logger.warning("Setting local station from Kernel Dataset")
            self.config.stations.from_dataset_dataframe(self.kernel_dataset.df)

    def validate(self):
        """apply all validators"""
        self.validate_processing()
        self.validate_decimation_scheme_and_dataset_compatability()
        self.memory_check()
        self.validate_save_fc_settings()

    def valid_decimations(self) -> List[AuroraDecimationLevel]:
        """
        Get the decimation levels that are valid.
        This is used when iterating over decimation levels in the processing.
        We do not want to try processing invalid levels (they will fail).

        Returns
        -------
        dec_levels: list
            Decimations from the config that are valid.
        """
        # identify valid rows of processing summary
        tmp = self.processing_summary[self.processing_summary.valid]
        valid_levels = tmp.dec_level.unique()

        dec_levels = [x for x in self.config.decimations]
        dec_levels = [x for x in dec_levels if x.decimation.level in valid_levels]
        msg = f"After validation there are {len(dec_levels)} valid decimation levels"
        logger.info(msg)
        return dec_levels

    def validate_save_fc_settings(self):
        """
        Update save_fc values in the config to be appropriate.
        - If all FCs exist, set save_fc to False.

        """
        if self.all_fcs_already_exist:
            msg = "FC Layer already exists -- forcing processing config save_fcs=False"
            logger.info(msg)
            for dec_level_config in self.config.decimations:
                # if dec_level_config.save_fcs:
                dec_level_config.save_fcs = False
        if self.config.stations.remote:
            save_any_fcs = np.array([x.save_fcs for x in self.config.decimations]).any()
            if save_any_fcs:
                msg = "\n Saving FCs for remote reference processing is not supported"
                msg = f"{msg} \n - To save FCs, process as single station, then you can use the FCs for RR processing"
                msg = f"{msg} \n - See issue #319 for details "
                msg = f"{msg} \n - forcing processing config save_fcs = False "
                logger.warning(msg)
                for dec_level_config in self.config.decimations:
                    dec_level_config.save_fcs = False

        return

    def get_mth5_file_open_mode(self) -> str:
        """check the mode of an open mth5 (read, write, append)"""
        if self.all_fcs_already_exist:
            return "r"
        elif self.config.decimations[0].save_fcs:
            return "a"
        else:
            return "r"

    def is_valid_dataset(self, row, i_dec) -> bool:
        """
        Given a row from the RunSummary, answer the question:
         "Will this decimation level yield a valid dataset?"

        Parameters
        ----------
        row: pandas.core.series.Series
            Row of the self._dataset_df (corresponding to a run that will be processed)
        i_dec: integer
            refers to decimation level

        Returns
        -------
        is_valid: Bool
            Whether the (run, decimation_level) pair associated with this row yields
            a valid dataset
        """
        # Query processing on survey, station, run, decimation
        cond1 = self.processing_summary.survey == row.survey
        cond2 = self.processing_summary.station == row.station
        cond3 = self.processing_summary.run == row.run
        cond4 = self.processing_summary.dec_level == i_dec
        cond5 = self.processing_summary.start == row.start

        cond = cond1 & cond2 & cond3 & cond4 & cond5
        processing_row = self.processing_summary[cond]
        if len(processing_row) != 1:
            return False
        is_valid = processing_row.valid.iloc[0]
        return is_valid

    @property
    def processing_type(self):
        """
        A description of the processing, will get passed to TF object,
        can be used for Z-file

        Could add a version or a hashtag to this
        Could also check dataset_df
        If remote.all==False append "Single Station"
        """
        processing_type = "Aurora"
        if self.dataset_df.remote.any():
            processing_type = f"{processing_type} Robust Remote Reference"
        else:
            processing_type = f"{processing_type} Robust Single Station"
        if "processing_type" in self.dataset_df.columns:
            processing_type = self.dataset_df["processing_type"].iloc[0]

        return processing_type

    def export_tf_collection(self, tf_collection: TransferFunctionCollection):
        """
        Assign transfer_function, residual_covariance, inverse_signal_power, station, survey

        Parameters
        ----------
        tf_collection: aurora.transfer_function.TransferFunctionCollection
            Contains TF estimates, covariance, and signal power values

        Returns
        -------
        tf_cls: mt_metadata.transfer_functions.core.TF
            Transfer function container
        """

        def make_decimation_dict_for_tf(
            tf_collection: TransferFunctionCollection,
            processing_config: Processing,
        ) -> dict:
            """
            Helper function to create a dictionary used by mt_metadata's TF class when
            writing z-files.  If no z-files will be written this is not needed

            sample element of decimation_dict:
            '1514.70134': {'level': 4, 'bands': (5, 6), 'npts': 386, 'df': 0.015625}}

            Note #1:  The line that does
            period_value["npts"] = tf_collection.tf_dict[i_dec].num_segments.data[0, i_band]
            doesn't feel very robust ... it would be better to explicitly select based on the num_segments
            xarray location where period==fb.center_period.  This is a placeholder for now.  In future,
            the TTFZ class is likely to be deprecated in favour of directly packing mt_metadata TFs during processing.

            Parameters
            ----------
            tf_collection: TransferFunctionCollection
                Collection of transfer function estimates from aurora.
            processing_config: Processing
                Instructions for processing with aurora

            Returns
            -------
            decimation_dict: dict
                Keyed by a string representing the period
                Values are a custom dictionary.
            """
            from mt_metadata.transfer_functions.io.zfiles.zmm import PERIOD_FORMAT

            decimation_dict = {}
            # dec_level_cfg is an AuroraDecimationLevel
            for i_dec, dec_level_cfg in enumerate(processing_config.decimations):
                for i_band, band in enumerate(dec_level_cfg.bands):
                    period_key = f"{band.center_period:{PERIOD_FORMAT}}"
                    period_value = {}
                    period_value["level"] = i_dec + 1  # +1 to match EMTF standard
                    period_value["bands"] = tuple(band.harmonic_indices[np.r_[0, -1]])
                    period_value["sample_rate"] = dec_level_cfg.decimation.sample_rate
                    try:
                        period_value["npts"] = tf_collection.tf_dict[
                            i_dec
                        ].num_segments.data[0, i_band]
                    except KeyError:
                        logger.warning(
                            f"Decimation level {i_dec} band {i_band} is invalid, not enough points."
                        )
                        period_value["npts"] = 0
                    decimation_dict[period_key] = period_value

            return decimation_dict

        channel_nomenclature = self.config.channel_nomenclature
        channel_nomenclature_dict = channel_nomenclature.to_dict()[
            "channel_nomenclature"
        ]
        merged_tf_dict = tf_collection.get_merged_dict(channel_nomenclature)
        decimation_dict = make_decimation_dict_for_tf(
            tf_collection, self.processing_config
        )

        tf_cls = TF(
            channel_nomenclature=channel_nomenclature_dict,
            decimation_dict=decimation_dict,
        )
        renamer_dict = {"output_channel": "output", "input_channel": "input"}
        tmp = merged_tf_dict["tf"].rename(renamer_dict)
        tf_cls.transfer_function = tmp

        isp = merged_tf_dict["cov_ss_inv"]
        renamer_dict = {"input_channel_1": "input", "input_channel_2": "output"}
        isp = isp.rename(renamer_dict)
        tf_cls.inverse_signal_power = isp

        res_cov = merged_tf_dict["cov_nn"]
        renamer_dict = {
            "output_channel_1": "input",
            "output_channel_2": "output",
        }
        res_cov = res_cov.rename(renamer_dict)
        tf_cls.residual_covariance = res_cov

        # Set survey metadata from the dataset
        # self.dataset.survey_metadata now returns a Survey object (not a dict)
        # Only set it if the TF object doesn't already have survey metadata
        # if tf_cls.survey_metadata is None or (
        #     hasattr(tf_cls.survey_metadata, "__len__")
        #     and len(tf_cls.survey_metadata) == 0
        # ):
        survey_obj = self.dataset.survey_metadata
        if survey_obj is not None:
            tf_cls.survey_metadata = survey_obj

        # Set station metadata and processing info
        tf_cls.station_metadata.provenance.creation_time = pd.Timestamp.now()
        tf_cls.station_metadata.provenance.processing_type = self.processing_type
        tf_cls.station_metadata.transfer_function.processed_date = pd.Timestamp.now()

        # Get runs processed from the dataset dataframe
        runs_processed = self.dataset_df.run.unique().tolist()
        tf_cls.station_metadata.transfer_function.runs_processed = runs_processed
        # TODO: tf_cls.station_metadata.transfer_function.processing_config = self.processing_config

        tf_cls.station_metadata.transfer_function.software.author = "K. Kappler"
        tf_cls.station_metadata.transfer_function.software.name = "Aurora"
        tf_cls.station_metadata.transfer_function.software.version = aurora_version

        # modify the run metadata to match the channel nomenclature, this should only be done if the
        # channels are different than the expected channel_nomenclature
        channels_named_incorrectly = False
        for ch in tf_cls.station_metadata.channels_recorded:
            if ch not in channel_nomenclature_dict.values():
                logger.warning(
                    f"Channel '{ch}' not found in channel_nomenclature_dict values"
                )
                logger.warning(
                    f"Available values: {list(channel_nomenclature_dict.values())}"
                )
                channels_named_incorrectly = True

        # This should be a last ditch effor to rename channels, the nomenclature should
        # propagate from the MTH5 through the processing to the TF object
        if channels_named_incorrectly:
            logger.info(
                "Modifying channel nomenclature in station metadata to match specified channel_nomenclature"
            )
            for i_run, run in enumerate(tf_cls.station_metadata.runs):
                for channel in run.channels:
                    new_ch = channel.copy()
                    default_component = channel.component
                    if default_component not in channel_nomenclature_dict:
                        logger.error(
                            f"Component '{default_component}' not found in channel_nomenclature_dict"
                        )
                        logger.error(
                            f"Available keys: {list(channel_nomenclature_dict.keys())}"
                        )
                        raise KeyError(
                            f"Component '{default_component}' not found in channel_nomenclature_dict. Available: {list(channel_nomenclature_dict.keys())}"
                        )
                    new_component = channel_nomenclature_dict[default_component]
                    new_ch.component = new_component
                    tf_cls.station_metadata.runs[i_run].remove_channel(
                        default_component
                    )
                    tf_cls.station_metadata.runs[i_run].add_channel(new_ch)

        return tf_cls

    def memory_check(self) -> None:
        """

        Checks if a RAM issue should be anticipated.

        Notes:
        Requires an estimate of available RAM, and an estimate of the dataset size
        Available RAM is taken from psutil,
        Dataset size is number of samples, times the number of bytes per sample
        Bits per sample is estimated to be 64 by default, (8-bytes)
        Returns
        -------

        """
        bytes_per_sample = 8  # 64-bit
        memory_threshold = 0.5

        # get the total amount of RAM in bytes
        total_memory = psutil.virtual_memory().total

        # print the total amount of RAM in GB
        logger.info(f"Total memory: {total_memory / (1024 ** 3):.2f} GB")
        num_samples = self.dataset_df.duration * self.dataset_df.sample_rate
        total_samples = num_samples.sum()
        total_bytes = total_samples * bytes_per_sample
        logger.info(f"Total Bytes of Raw Data: {total_bytes / (1024 ** 3):.3f} GB")

        ram_fraction = 1.0 * total_bytes / total_memory
        logger.info(f"Raw Data will use: {100 * ram_fraction:.3f} % of memory")

        # Check a condition
        if total_bytes > memory_threshold * total_memory:
            self._memory_warning = True
        else:
            self._memory_warning = False

        if self._memory_warning:
            raise Exception("Job requires too much memory")


# ###################################################################
# ######## Helper Functions #########################################
# ###################################################################


def _decimation_level_info_valid(
    df: pd.DataFrame,
) -> bool:
    """
        Applies some sanity checks to a processing summary dataframe chunk.

        Development Notes:
        This was previously a fairly janky block of code in `update_processing_summary`.
        It is being factored to label and contain the jank, and could still be improved.
        In particular, just because the conditions being checked are expected, it
        may be that they do not make the processing invalid if they are not True.
        i.e. The conditions being checked are likely overly conservative and can be relaxed.


        Parameters
        ----------
        df: pd.DataFrame
            Input has already been grouped by ["survey", "station", "run", "start",]
            This is like a "run chunk".
            This run chunk can have several decimation levels, often these are
            [1, 4, 4, 4]

    Returns
    -------

    """
    # assert that decimation levels are incrementing (usually 0,1,2,3...)
    cond = (df.dec_level.diff()[1:] == 1).all()
    if cond:  # dec levels increment by 1
        pass
    else:
        msg = "Processing summary chunk decimation levels are not incrementing by 1."
        logger.error(msg)
        return False

    # check that first decimation level is zero
    cond1 = df.dec_factor.iloc[0] == 1
    cond2 = df.dec_level.iloc[0] == 0
    cond = cond1 & cond2
    if cond:
        pass
    else:
        msg = "Processing summary chunk was expecting first decimation level 0 and decimation factor 1"
        logger.error(msg)
        return False

    return True


def _make_processing_summary(
    kernel_dataset_df: pd.DataFrame,
    decimation_info: dict,
    window_params_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Creates or updates the processing summary table based on processing parameters
    and kernel dataset.
    - Melt the decimation levels over the run summary.
    - Add columns to estimate the number of FFT windows for each row

    Returns
    -------
    processing_summary: pd.DataFrame
        One row per each run-decimation pair
    """

    # Create correctly shaped dataframe
    tmp = kernel_dataset_df.copy(deep=True)
    id_vars = list(tmp.columns)
    for i_dec, dec_factor in decimation_info.items():
        tmp[i_dec] = dec_factor
    tmp = tmp.melt(id_vars=id_vars, value_name="dec_factor", var_name="dec_level")
    sortby = ["survey", "station", "run", "start", "dec_level"]
    tmp.sort_values(by=sortby, inplace=True)
    tmp.reset_index(drop=True, inplace=True)
    tmp.drop("sample_rate", axis=1, inplace=True)  # not valid for decimated data

    # Add window info
    group_by = [
        "survey",
        "station",
        "run",
        "start",
    ]
    chunks = []
    grouper = tmp.groupby(group_by)

    for group, chunk_df in grouper:
        if not _decimation_level_info_valid(chunk_df):
            msg = f"Failed to update processing summary -- {group} had invalid decimation level info"
            logger.error(msg)
            return
        # df.sample_rate /= np.cumprod(df.dec_factor)  # better to take from config, maybe add as a test though?
        chunk_df.reset_index(inplace=True, drop=True)
        chunk_df = chunk_df.join(window_params_df)
        chunk_df["num_samples"] = np.floor(chunk_df.duration * chunk_df.sample_rate)
        num_windows = np.zeros(len(chunk_df))
        for i, row in chunk_df.iterrows():
            ws = WindowingScheme(
                num_samples_window=row.num_samples_window,
                num_samples_overlap=row.num_samples_overlap,
            )
            num_windows[i] = ws.available_number_of_windows(row.num_samples)
        chunk_df["num_stft_windows"] = num_windows
        chunks.append(chunk_df)

    processing_summary = pd.concat(chunks)
    processing_summary.reset_index(drop=True, inplace=True)
    return processing_summary


@path_or_mth5_object
def mth5_has_fcs(m, survey_id, station_id, run_id, remote, processing_config, **kwargs):
    """
    Checks if all needed fc-levels for survey-station-run are present under processing_config

    Note #1: At this point in the logic, it is established that there are FCs associated with run_id and there are
        at least as many FC decimation levels as we require as per the processing config.  The next step is to
        assert whether it is True that the existing FCs conform to the recipe in the processing config.

    kwargs are here as a pass through to the decorator ... we pass mode="r","a","w"

    Parameters
    ----------
    m
    survey_id
    station_id
    run_id
    dataset_df

    Returns
    -------

    """
    row_ssr_str = f"survey: {survey_id}, station: {station_id}, run: {run_id}"
    survey_obj = m.get_survey(survey_id)
    station_obj = survey_obj.stations_group.get_station(station_id)
    if not station_obj.fourier_coefficients_group.groups_list:
        msg = f"Fourier coefficients not detected for {row_ssr_str}"
        msg += "-- Fourier coefficients will be computed"
        logger.info(msg)
        return False

    logger.info("FCs detected -- checking against processing requirements.")
    try:
        fc_group = station_obj.fourier_coefficients_group.get_fc_group(run_id)
    except MTH5Error:
        msg = f"Run {run_id} not found in FC Groups -- need to compute FCs"
        logger.info(msg)
        return False

    if len(fc_group.groups_list) < processing_config.num_decimation_levels:
        msg = f"Not enough FC Groups found for {row_ssr_str} -- will build them"
        return False

    # Can check time periods here if desired, but unique (survey, station, run) should make this unneeded
    # processing_run = self.processing_config.stations.local.get_run(run_id)
    # for tp in processing_run.time_periods:
    #    assert tp in fc_group time periods

    # See note #1
    fcs_already_there = fc_group.supports_aurora_processing_config(
        processing_config, remote
    )
    return fcs_already_there


# =============================================================================
# Helper Functions
# =============================================================================


def station_obj_from_row(row):
    """
    Access the station object
    Note if/else could avoidable if replacing text string "none" with a None object in survey column

    Parameters
    ----------
    row: pd.Series
        A row of tfk.dataset_df

    Returns
    -------
    station_obj:

    """
    if row.mth5_obj.file_version == "0.1.0":
        station_obj = row.mth5_obj.stations_group.get_station(row.station)
    elif row.mth5_obj.file_version == "0.2.0":
        survey_group = row.mth5_obj.surveys_group.get_survey(row.survey)
        station_obj = survey_group.stations_group.get_station(row.station)
    return station_obj



"""
This module contains a base class for Transfer function information.

Development Notes:
 The class is based on Gary's TTF.m in iris_mt_scratch egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes

TODO: This class should be updated to use mt_metadata.transfer_function.TF, or replaced by it (issue #203)

"""

import numpy as np
import xarray as xr
from aurora.config.metadata.processing import Processing
from loguru import logger
from mt_metadata.processing.aurora import FrequencyBands
from typing import Optional, Union


class TransferFunction:
    """
    Class to contain transfer function array.

    Parameters
    ----------
    TF : numpy array
        array of transfer functions: TF(Nout, Nin, Nperiods)
    T : numpy array
        list of periods
    cov_ss_inv : numpy array
        inverse signal power matrix.  aka Cov_SS in EMTF matlab codes
    cov_nn : numpy array
        noise covariance matrix: aka Cov_NN in EMTF matlab codes
    num_segments : integer array?
        Number of samples used to estimate TF for each band, and for each \
        output channel (might be different for different channels)
    R2 : xarray.DataArray
        multiple coherence for each output channel / band
    FullCov : boolean
        true if full covariance is provided

    """

    def __init__(
        self,
        decimation_level_id: int,
        frequency_bands: FrequencyBands,
        processing_config: Optional[Union[Processing, None]] = None,
        **kwargs
    ):
        """
        Constructor.

        Parameters
        ----------
        _emtf_header : legacy header information used by Egbert's matlab class.  Header contains
        local site header, remote site header if appropriate, and information about estimation approach
        decimation_level_id: int
            Identifies the relevant decimation level.  Used for accessing the
            appropriate info in self.processing config.
        frequency_bands: FrequencyBands
            frequency bands object defining the tf estimation bands.
        """
        self._emtf_tf_header = None
        self.decimation_level_id = decimation_level_id
        self.frequency_bands = frequency_bands
        self.num_segments = None
        self.cov_ss_inv = None
        self.cov_nn = None
        self.R2 = None
        self.initialized = False
        self.processing_config = processing_config

        if self.emtf_tf_header is not None:
            if self.num_bands is not None:
                self._initialize_arrays()

    @property
    def emtf_tf_header(self):
        """

        Returns
        -------
        emtf_tf_header: None or OrderedDict
            A legacy construction from the old EMTF codes with some metadata about the TF.
        """
        if self.processing_config is None:
            logger.info("No header is available without a processing config")
            self._emtf_tf_header = None
        else:
            if self._emtf_tf_header is None:
                tfh = self.processing_config.emtf_tf_header(self.decimation_level_id)
                self._emtf_tf_header = tfh
        return self._emtf_tf_header

    @property
    def tf_header(self):
        """returns the emtf_header"""
        return self.emtf_tf_header

    @property
    def tf(self):
        return self.transfer_function

    @property
    def num_bands(self) -> int:
        """
        Get number of bands associated with TF

        Returns
        -------
        num_bands: int
            a count of the frequency bands associated with the TF
        """
        return self.frequency_bands.number_of_bands

    @property
    def periods(self):
        periods = self.frequency_bands.band_centers(frequency_or_period="period")
        periods = np.flipud(periods)
        return periods

    def _initialize_arrays(self) -> None:
        """
        Initialize the xarray objects that will contain the TF data.

        Development Notes:
        There are four separate data strucutres, indexed by period here:
        - TF (num_channels_out, num_channels_in)
        - cov_ss_inv (num_channels_in, num_channels_in, num_bands),
        - Cov_NN (num_channels_out, num_channels_out),
        - R2 num_channels_out)

        We use frequency in Fourier domain and period in TF domain.  Might
        be better to be consistent.  It would be inexpensive to duplicate
        the axis and simply provide both.
        """
        if self.tf_header is None:
            logger.error("header needed to allocate transfer function arrays")
            raise Exception

        # <transfer function xarray>
        tf_array_dims = (self.num_channels_out, self.num_channels_in, self.num_bands)
        tf_array = np.zeros(tf_array_dims, dtype=np.complex128)
        self.transfer_function = xr.DataArray(
            tf_array,
            dims=["output_channel", "input_channel", "period"],  # frequency"],
            coords={
                "period": self.periods,
                "output_channel": self.tf_header.output_channels,
                "input_channel": self.tf_header.input_channels,
            },
        )

        # num_segments xarray
        num_segments = np.zeros((self.num_channels_out, self.num_bands), dtype=np.int32)
        num_segments_xr = xr.DataArray(
            num_segments,
            dims=["channel", "period"],
            coords={
                "period": self.periods,
                "channel": self.tf_header.output_channels,
            },
        )
        self.num_segments = num_segments_xr

        # Inverse signal covariance
        cov_ss_dims = (self.num_channels_in, self.num_channels_in, self.num_bands)
        cov_ss_inv = np.zeros(cov_ss_dims, dtype=np.complex128)
        self.cov_ss_inv = xr.DataArray(
            cov_ss_inv,
            dims=["input_channel_1", "input_channel_2", "period"],
            coords={
                "input_channel_1": self.tf_header.input_channels,
                "input_channel_2": self.tf_header.input_channels,
                "period": self.periods,
            },
        )

        # Noise covariance
        cov_nn_dims = (self.num_channels_out, self.num_channels_out, self.num_bands)
        cov_nn = np.zeros(cov_nn_dims, dtype=np.complex128)
        self.cov_nn = xr.DataArray(
            cov_nn,
            dims=["output_channel_1", "output_channel_2", "period"],
            coords={
                "output_channel_1": self.tf_header.output_channels,
                "output_channel_2": self.tf_header.output_channels,
                "period": self.periods,
            },
        )

        # Coefficient of determination
        self.R2 = xr.DataArray(
            np.zeros((self.num_channels_out, self.num_bands)),
            dims=["output_channel", "period"],
            coords={
                "output_channel": self.tf_header.output_channels,
                "period": self.periods,
            },
        )

        self.initialized = True

    @property
    def minimum_period(self) -> float:
        """return the minimum (shortest) period in self.periods"""
        return np.min(self.periods)

    @property
    def maximum_period(self):
        """return the maximum (longest) period in self.periods"""
        return np.max(self.periods)

    @property
    def num_channels_in(self):
        """return the number of input channels in the TF"""
        return len(self.tf_header.input_channels)

    @property
    def num_channels_out(self):
        """return the number of output channels in the TF"""
        return len(self.tf_header.output_channels)

    def frequency_index(self, frequency):
        """
        Gets the index of the tf array associated with the input frequency

        Parameters
        ----------
        frequency: float
            The period in seconds

        Returns
        -------
        frequency_index: int
            The index of the tf array corresponding to the frequency
        """
        return self.period_index(1.0 / frequency)

    def period_index(self, period: float) -> int:
        """
        Gets the index of the tf array associated with the input period.

        Parameters
        ----------
        period: float
            The period in seconds

        Returns
        -------
        period_index: int
            The index of the tf array corresponding to the period
        """
        period_index = np.isclose(self.num_segments.period, period)
        period_index = np.where(period_index)[0][0]
        return period_index

    def set_tf(self, regression_estimator, period):
        """
        Sets TF elements for one band, using contents of regression_estimator object.

        """
        index = self.period_index(period)

        tf = regression_estimator.b_to_xarray()
        output_channels = list(regression_estimator._Y.data_vars)
        input_channels = list(regression_estimator._X.data_vars)

        for out_ch in output_channels:
            for inp_ch in input_channels:
                self.tf[:, :, index].loc[out_ch, inp_ch] = tf.loc[out_ch, inp_ch]

        if regression_estimator.noise_covariance is not None:
            for out_ch_1 in output_channels:
                for out_ch_2 in output_channels:
                    tmp = regression_estimator.cov_nn.loc[out_ch_1, out_ch_2]
                    self.cov_nn[:, :, index].loc[out_ch_1, out_ch_2] = tmp

        if regression_estimator.inverse_signal_covariance is not None:
            for inp_ch_1 in input_channels:
                for inp_ch_2 in input_channels:
                    tmp = regression_estimator.cov_ss_inv.loc[inp_ch_1, inp_ch_2]
                    self.cov_ss_inv[:, :, index].loc[inp_ch_1, inp_ch_2] = tmp

        if regression_estimator.R2 is not None:
            for out_ch in output_channels:
                tmp = regression_estimator.R2.loc[out_ch]
                self.R2[:, index].loc[out_ch] = tmp

        self.num_segments.data[:, index] = regression_estimator.n_data

        return



"""
This module contains aurora methods associated with spectrograms or "STFTs".
In future these tools should be moved to MTH5 and made methods of the Spectrogram class.
For now, we can use this module as a place to aggregate functions to migrate.
"""

from aurora.config.metadata.processing import Processing as AuroraProcessing
from aurora.pipelines.time_series_helpers import truncate_to_clock_zero
from aurora.pipelines.transfer_function_kernel import station_obj_from_row
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.sandbox.triage_metadata import triage_run_id
from aurora.time_series.xarray_helpers import nan_to_mean
from aurora.time_series.xarray_helpers import time_axis_match
from aurora.time_series.windowed_time_series import WindowedTimeSeries
from aurora.time_series.windowing_scheme import window_scheme_from_decimation
from loguru import logger
from mt_metadata.processing.aurora import DecimationLevel as AuroraDecimationLevel
from mth5.groups import RunGroup
from mth5.processing.spectre.prewhitening import apply_prewhitening
from mth5.processing.spectre.prewhitening import apply_recoloring
from typing import Literal, Optional

import mth5.timeseries.spectre as spectre
import numpy as np
import pandas as pd
import xarray as xr


def make_stft_objects(
    processing_config: AuroraProcessing,
    i_dec_level: int,
    run_obj: RunGroup,
    run_xrds: xr.Dataset,
    units: Literal["MT", "SI"] = "MT",
) -> xr.Dataset:
    """
    Applies STFT to all channel time series in the input run.

    This method could be modified in a multiple station code so that it doesn't care
    if the station is "local" or "remote" but rather uses scale factors keyed by
    station_id (WIP - issue #329)

    Parameters
    ----------
    processing_config: mt_metadata.processing.aurora.Processing
        Metadata about the processing to be applied
    i_dec_level: int
        The decimation level to process
    run_obj: mth5.groups.master_station_run_channel.RunGroup
        The run to transform to stft
    run_xrds: xarray.core.dataset.Dataset
        The data time series from the run to transform
    units: Literal["MT", "SI"] = "MT",
        expects "MT".  May change so that this is the only accepted set of units

    Returns
    -------
    stft_obj: xarray.core.dataset.Dataset
        Time series of calibrated Fourier coefficients per each channel in the run

    Development Notes:
    Here are the parameters that are defined via the mt_metadata fourier coefficients structures:

    "bands",
    "decimation.anti_alias_filter": "default",
    "decimation.factor": 4.0,
    "decimation.level": 2,
    "decimation.method": "default",
    "decimation.sample_rate": 0.0625,
    "stft.per_window_detrend_type": "linear",
    "stft.prewhitening_type": "first difference",
    "stft.window.clock_zero_type": "ignore",
    "stft.window.num_samples": 128,
    "stft.window.overlap": 32,
    "stft.window.type": "boxcar"

    Creating the decimations config requires a decision about decimation factors and the number of levels.
    We have been getting this from the EMTF band setup file by default.  It is desirable to continue supporting this,
    however, note that the EMTF band setup is really about a time series operation, and not about making STFTs.

    For the record, here is the legacy decimation config from EMTF, a.k.a. decset.cfg:
    ```
    4     0      # of decimation level, & decimation offset
    128  32.   1   0   0   7   4   32   1
    1.0
    128  32.   4   0   0   7   4   32   4
    .2154  .1911   .1307   .0705
    128  32.   4   0   0   7   4   32   4
    .2154  .1911   .1307   .0705
    128  32.   4   0   0   7   4   32   4
    .2154  .1911   .1307   .0705
    ```

    This essentially corresponds to a "Decimations Group" which is a list of decimations.
    Related to the generation of FCs is the ARMA prewhitening (Issue #60) which was controlled in
    EMTF with pwset.cfg
    4    5             # of decimation levels, # of channels
    3 3 3 3 3
    3 3 3 3 3
    3 3 3 3 3
    3 3 3 3 3

    """
    stft_config = processing_config.get_decimation_level(i_dec_level)
    spectrogram = run_ts_to_stft(stft_config, run_xrds)
    run_id = run_obj.metadata.id
    if run_obj.station_metadata.id == processing_config.stations.local.id:
        scale_factors = processing_config.stations.local.run_dict[
            run_id
        ].channel_scale_factors
    elif run_obj.station_metadata.id == processing_config.stations.remote[0].id:
        scale_factors = (
            processing_config.stations.remote[0].run_dict[run_id].channel_scale_factors
        )

    stft_obj = calibrate_stft_obj(
        spectrogram.dataset,
        run_obj,
        units=units,
        channel_scale_factors=scale_factors,
    )

    return stft_obj


def shape_check_spectrograms(local_stfts: list, remote_stfts: list):
    """
    Performs a shape check on the local and remote STFTs.
    Seems associated with getting one fewer sample than expected from the edge of a run,
    WIP: Aurora Issue #289.

    If a shape mismatch is detected, compute the greatest lower bound (glb) and least upper bound (lub)
    of the time axis of the local and remote STFTs, and then trim both STFTs to the intersection of these bounds.

    Parameters
    ----------
    local_stfts: list
        A list of local STFT objects, each is an xr.Dataset
    remote_stfts: list
        A list of remote STFT objects, each is an xr.Dataset

    returns: Tuple
        (local_stfts, remote_stfts)
        The original input arguments, shape-matched

    """
    n_chunks = len(local_stfts)
    for i_chunk in range(n_chunks):
        ok = local_stfts[i_chunk].time.shape == remote_stfts[i_chunk].time.shape
        if not ok:
            logger.warning("Mismatch in FC array lengths detected -- Issue #289")
            glb = max(
                local_stfts[i_chunk].time.min(),
                remote_stfts[i_chunk].time.min(),
            )
            lub = min(
                local_stfts[i_chunk].time.max(),
                remote_stfts[i_chunk].time.max(),
            )

            cond1 = local_stfts[i_chunk].time >= glb
            cond2 = local_stfts[i_chunk].time <= lub
            local_stfts[i_chunk] = local_stfts[i_chunk].where(cond1 & cond2, drop=True)

            cond1 = remote_stfts[i_chunk].time >= glb
            cond2 = remote_stfts[i_chunk].time <= lub
            remote_stfts[i_chunk] = remote_stfts[i_chunk].where(
                cond1 & cond2, drop=True
            )
            assert local_stfts[i_chunk].time.shape == remote_stfts[i_chunk].time.shape

    return local_stfts, remote_stfts


def time_axis_check_spectrograms(local_stfts: list, remote_stfts: list):
    """
    Checks that the time axes of local and remote STFTs match.
    If they do not match, raises an error.

    Parameters
    ----------
    local_stfts: list
        A list of local STFT objects, each is an xr.Dataset
    remote_stfts: list
        A list of remote STFT objects, each is an xr.Dataset

    Raises
    ------
    ValueError
        If the time axes of local and remote STFTs do not match.
    """
    for i_chunk in range(len(local_stfts)):
        ok = time_axis_match(local_stfts[i_chunk], remote_stfts[i_chunk])
        if not ok:
            raise ValueError(
                "Time axes of local and remote STFTs do not match. "
                "Please check the data."
            )


def merge_stfts(stfts: dict, tfk: TransferFunctionKernel):
    """

    Applies concatenation along the time axis to multiple arrays of STFTs from different runs.
    At the TF estimation level we treat all the FCs in one array.
    This builds the array for both the local and the remote STFTs.

    Parameters
    ----------
    stfts: dict
        The dict is keyed by "local" and "remote".
        Each value is a list of STFTs (one list for local and one for remote)
    tfk: TransferFunctionKernel
        Just here to let us know if there is a remote reference to merge or not.

    Returns
    -------
    local_merged_stft_obj, remote_merged_stft_obj: Tuple
        Both are xr.Datasets
    """
    local_stfts = stfts["local"]
    remote_stfts = stfts["remote"]
    if tfk.config.stations.remote:
        local_stfts, remote_stfts = shape_check_spectrograms(local_stfts, remote_stfts)
        time_axis_check_spectrograms(local_stfts, remote_stfts)
        remote_merged_stft_obj = xr.concat(remote_stfts, "time")
    else:
        remote_merged_stft_obj = None

    local_merged_stft_obj = xr.concat(local_stfts, "time")

    return local_merged_stft_obj, remote_merged_stft_obj


def append_chunk_to_stfts(stfts: dict, chunk: xr.Dataset, remote: bool) -> dict:
    """
    Aggregate one STFT into a larger dictionary that tracks all the STFTs

    Parameters
    ----------
    stfts: dict
        has keys "local" and "remote".
    chunk: xr.Dataset
        The data to append to the dictionary
    remote: bool
        If True, append the chunk to stfts["remote"], else append to stfts["local"]

    Returns
    -------
    stfts: dict
        Same as input but now has new chunk appended to it.
    """
    if remote:
        stfts["remote"].append(chunk)
    else:
        stfts["local"].append(chunk)
    return stfts


# def load_spectrogram_from_station_object(station_obj, fc_group_id, fc_decimation_id):
#     """
#     Placeholder.  This could also be a method in mth5
#     Returns
#     -------
#
#     """
#     return station_obj.fourier_coefficients_group.get_fc_group(fc_group_id).get_decimation_level(fc_decimation_id)


def load_stft_obj_from_mth5(
    i_dec_level: int,
    row: pd.Series,
    run_obj: RunGroup,
    channels: Optional[list] = None,
) -> xr.Dataset:
    """
    Load stft_obj from mth5 (instead of compute)

    Note #1: See note #1 in mth5.timeseries.spectre.spectrogram.py in extract_band function.

    Parameters
    ----------
    i_dec_level: int
        The decimation level where the data are stored within the Fourier Coefficient group
    row: pandas.core.series.Series
        A row of the TFK.dataset_df
    run_obj: mth5.groups.run.RunGroup
        The original time-domain run associated with the data to load

    Returns
    -------
    stft_chunk: xr.Dataset
        An STFT from mth5.
    """
    station_obj = station_obj_from_row(row)
    fc_group = station_obj.fourier_coefficients_group.get_fc_group(run_obj.metadata.id)
    fc_decimation_level = fc_group.get_decimation_level(f"{i_dec_level}")
    stft_obj = fc_decimation_level.to_xarray(channels=channels)

    cond1 = stft_obj.time >= row.start.tz_localize(None)
    cond2 = stft_obj.time <= row.end.tz_localize(None)
    try:
        stft_chunk = stft_obj.where(cond1 & cond2, drop=True)
    except TypeError:  # see Note #1
        tmp = stft_obj.to_array()
        tmp = tmp.where(cond1 & cond2, drop=True)
        stft_chunk = tmp.to_dataset("variable")
    return stft_chunk


def save_fourier_coefficients(
    dec_level_config: AuroraDecimationLevel, row: pd.Series, run_obj, stft_obj
) -> None:
    """
    Optionally saves the stft object into the MTH5.
    Note that the dec_level_config must have its save_fcs attr set to True to actually save the data.
    WIP

    Note #1: Logic for building FC layers:
    If the processing config decimation_level.save_fcs_type = "h5" and fc_levels_already_exist is False, then open
    in append mode, else open in read mode.  We should support a flag: force_rebuild_fcs, normally False.  This flag
    is only needed when save_fcs_type=="h5".  If True, then we open in append mode, regarless of fc_levels_already_exist
    The task of setting mode="a", mode="r" can be handled by tfk (maybe in tfk.validate())

    Parameters
    ----------
    dec_level_config: mt_metadata.processing.aurora.decimation_level.DecimationLevel
        The information about decimation level associated with row, run, stft_obj
    row: pd.Series
         A row of the TFK.dataset_df
    run_obj: mth5.groups.run.RunGroup
        The run object associated with the STFTs.
    stft_obj: xr.Dataset
        The data to pack intp the mth5.

    """

    if not dec_level_config.save_fcs:
        msg = "Skip saving FCs. dec_level_config.save_fc = "
        msg = f"{msg} {dec_level_config.save_fcs}"
        logger.info(f"{msg}")
        return
    i_dec_level = dec_level_config.decimation.level
    if dec_level_config.save_fcs_type == "csv":
        msg = "Unless debugging or testing, saving FCs to csv unexpected"
        logger.warning(msg)
        csv_name = f"{row.station}_dec_level_{i_dec_level}.csv"
        stft_df = stft_obj.to_dataframe()
        stft_df.to_csv(csv_name)
    elif dec_level_config.save_fcs_type == "h5":
        logger.info(("Saving FC level"))
        station_obj = station_obj_from_row(row)

        if not row.mth5_obj.h5_is_write():
            msg = "See Note #1: Logic for building FC layers"
            raise NotImplementedError(msg)

        # Get FC group (create if needed)
        if run_obj.metadata.id in station_obj.fourier_coefficients_group.groups_list:
            fc_group = station_obj.fourier_coefficients_group.get_fc_group(
                run_obj.metadata.id
            )
        else:
            fc_group = station_obj.fourier_coefficients_group.add_fc_group(
                run_obj.metadata.id
            )

        decimation_level_metadata = dec_level_config.to_fc_decimation()

        # Get FC Decimation Level (create if needed)
        dec_level_name = f"{i_dec_level}"
        if dec_level_name in fc_group.groups_list:
            fc_decimation_level = fc_group.get_decimation_level(dec_level_name)
            fc_decimation_level.metadata = decimation_level_metadata
        else:
            fc_decimation_level = fc_group.add_decimation_level(
                dec_level_name,
                decimation_level_metadata=decimation_level_metadata,
            )
        fc_decimation_level.from_xarray(
            stft_obj, decimation_level_metadata.decimation.sample_rate
        )
        fc_decimation_level.update_metadata()
        fc_group.update_metadata()
    else:
        msg = f"dec_level_config.save_fcs_type {dec_level_config.save_fcs_type} not recognized"
        msg = f"{msg} \n Skipping save fcs"
        logger.warning(msg)
    return


def get_spectrograms(
    tfk: TransferFunctionKernel, i_dec_level: int, units: str = "MT"
) -> dict:
    """
    Given a decimation level id, loads a dictianary of all spectragrams from information in tfk.
    TODO: Make this a method of TFK
    TODO: Modify this to be able to yield Spectrogram objects.

    Parameters
    ----------
    tfk: TransferFunctionKernel

    i_dec_level: integer
        The decimation level of the spectrograms.

    units: str
        "MT" or "SI", likely to be deprecated

    Returns
    -------
    stfts: dict
        The short time fourier transforms for the decimation level as a dictionary.
        Keys are "local" and "remote".  Values are lists, one (element) xr.Dataset per run
    """

    stfts = {}
    stfts["local"] = []
    stfts["remote"] = []

    # Check first if TS processing or accessing FC Levels
    for i, row in tfk.dataset_df.iterrows():
        # TODO: Consider updating this to iterate over row-pairs corresponding to simultaneous data.
        #  Grouping by (sorted) starttime should accomplish this.

        if not tfk.is_valid_dataset(row, i_dec_level):
            continue

        run_obj = row.mth5_obj.from_reference(row.run_hdf5_reference)
        if row.fc:
            stft_obj = load_stft_obj_from_mth5(i_dec_level, row, run_obj)
            # TODO: Cast stft_obj to a Spectrogram here
            stfts = append_chunk_to_stfts(stfts, stft_obj, row.remote)
            continue

        run_xrds = row["run_dataarray"].to_dataset("channel")

        # Musgraves workaround for old MT data
        triage_run_id(row.run, run_obj)

        stft_obj = make_stft_objects(
            tfk.config,
            i_dec_level,
            run_obj,
            run_xrds,
            units,
        )
        # TODO: Cast stft_obj to a Spectrogram here or in make_stft_objects

        # Pack FCs into h5
        dec_level_config = tfk.config.decimations[i_dec_level]
        save_fourier_coefficients(dec_level_config, row, run_obj, stft_obj)
        # TODO: cast stft_obj to a Spectrogram here

        stfts = append_chunk_to_stfts(stfts, stft_obj, row.remote)

    return stfts


def run_ts_to_stft(
    decimation_obj: AuroraDecimationLevel, run_xrds_orig: xr.Dataset
) -> spectre.Spectrogram:
    """
    Converts a runts object into a time series of Fourier coefficients.
    Similar to run_ts_to_stft_scipy, but in this implementation operations on individual
    windows are possible (for example pre-whitening per time window via ARMA filtering).

    TODO: Make the output of this function a Spectrogram object

    Parameters
    ----------
    decimation_obj : AuroraDecimationLevel
        Information about how the decimation level is to be processed
    run_xrds_orig: xarray.core.dataset.Dataset
        normally extracted from mth5.RunTS

    Returns
    -------
    stft_obj: xarray.core.dataset.Dataset
        Note that the STFT object may have inf/nan in the DC harmonic, introduced by
        recoloring. This really doesn't matter since we don't use the DC harmonic for
        anything.
    """
    # need to remove any nans before windowing, or else if there is a single
    # nan then the whole channel becomes nan.
    run_xrds = nan_to_mean(run_xrds_orig)
    run_xrds = apply_prewhitening(decimation_obj.stft.prewhitening_type, run_xrds)
    run_xrds = truncate_to_clock_zero(decimation_obj, run_xrds)
    windowing_scheme = window_scheme_from_decimation(decimation_obj)
    windowed_obj = windowing_scheme.apply_sliding_window(
        run_xrds, dt=1.0 / decimation_obj.decimation.sample_rate
    )
    if not np.prod(windowed_obj.to_array().data.shape):
        raise ValueError

    windowed_obj = WindowedTimeSeries.detrend(data=windowed_obj, detrend_type="linear")
    tapered_obj = WindowedTimeSeries.apply_taper(
        data=windowed_obj, taper=windowing_scheme.taper
    )
    stft_obj = WindowedTimeSeries.apply_fft(
        data=tapered_obj,
        sample_rate=windowing_scheme.sample_rate,
        spectral_density_correction=windowing_scheme.linear_spectral_density_calibration_factor,
        detrend_type=decimation_obj.stft.per_window_detrend_type,
    )

    if decimation_obj.stft.recoloring:
        stft_obj = apply_recoloring(decimation_obj.stft.prewhitening_type, stft_obj)

    spectrogram = spectre.Spectrogram(dataset=stft_obj)
    return spectrogram


def calibrate_stft_obj(
    stft_obj: xr.Dataset,
    run_obj: RunGroup,
    units: Literal["MT", "SI"] = "MT",
    channel_scale_factors: Optional[dict] = None,
) -> xr.Dataset:
    """
    Calibrates frequency domain data into MT units.

    Development Notes:
     The calibration often raises a runtime warning due to DC term in calibration response = 0.
     TODO: It would be nice to suppress this, maybe by only calibrating the non-dc terms and directly assigning np.nan to the dc component when DC-response is zero.

    Parameters
    ----------
    stft_obj : xarray.core.dataset.Dataset
        Time series of Fourier coefficients to be calibrated
    run_obj : mth5.groups.master_station_run_channel.RunGroup
        Provides information about filters for calibration
    units : string
        usually "MT", contemplating supporting "SI"
    channel_scale_factors : Optional[dict]
        Keyed by channel, supports a single scalar to apply to that channels data
        Useful for debugging.  Should not be used in production and should throw a
        warning if it is not None

    Returns
    -------
    stft_obj : xarray.core.dataset.Dataset
        Time series of calibrated Fourier coefficients
    """
    for channel_id in stft_obj.keys():

        channel = run_obj.get_channel(channel_id)
        channel_response = channel.channel_response
        if not channel_response.filters_list:
            msg = f"Channel {channel_id} with empty filters list detected"
            logger.warning(msg)
            if channel_id == "hy":
                msg = "Channel hy has no filters, try using filters from hx"
                logger.warning(msg)
                channel_response = run_obj.get_channel("hx").channel_response

        indices_to_flip = channel_response.get_indices_of_filters_to_remove(
            include_decimation=False, include_delay=False
        )
        indices_to_flip = [
            i for i in indices_to_flip if channel.metadata.filters[i].applied
        ]
        filters_to_remove = [channel_response.filters_list[i] for i in indices_to_flip]
        if not filters_to_remove:
            logger.warning("No filters to remove")

        calibration_response = channel_response.complex_response(
            stft_obj.frequency.data, filters_list=filters_to_remove
        )

        if channel_scale_factors:
            try:
                channel_scale_factor = channel_scale_factors[channel_id]
            except KeyError:
                channel_scale_factor = 1.0
            if channel_scale_factor != 1.0:
                calibration_response /= channel_scale_factor

        if units == "SI":
            logger.warning("Warning: SI Units are not robustly supported issue #36")

        # Handle case where DC term in calibration response = 0 (Sometimes raises a runtime warning)
        zero_dc_term = calibration_response[0] == 0
        if zero_dc_term:
            calibration_response[0] = 1.0
        stft_obj[channel_id].data /= calibration_response
        if zero_dc_term:
            stft_obj[channel_id].data[:, 0] = np.nan + 1j * np.nan

    return stft_obj



"""Pytest translation of test_define_frequency_bands.py"""


from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.test_utils.synthetic.processing_helpers import get_example_kernel_dataset
from aurora.test_utils.synthetic.triage import tfs_nearly_equal


def test_can_declare_frequencies_directly_in_config(synthetic_test_paths):
    """Test that manually declared frequency bands produce same results as defaults.

    This test verifies that explicitly passing band_edges to create_from_kernel_dataset
    produces the same transfer function as using the default band setup. The key is to
    use the same num_samples_window in both configs, since band edges are calculated
    based on FFT harmonics which depend on the window size.
    """
    kernel_dataset = get_example_kernel_dataset()
    cc = ConfigCreator()
    cfg1 = cc.create_from_kernel_dataset(kernel_dataset, estimator={"engine": "RME"})
    decimation_factors = list(cfg1.decimation_info.values())
    band_edges = cfg1.band_edges_dict

    # Use the same num_samples_window as cfg1 (default is 256)
    # to ensure band_edges align with FFT harmonics
    num_samples_window = cfg1.decimations[0].stft.window.num_samples

    cfg2 = cc.create_from_kernel_dataset(
        kernel_dataset,
        estimator={"engine": "RME"},
        band_edges=band_edges,
        decimation_factors=decimation_factors,
        num_samples_window=len(band_edges) * [num_samples_window],
    )

    cfg1_path = synthetic_test_paths.aurora_results_path.joinpath("cfg1.xml")
    cfg2_path = synthetic_test_paths.aurora_results_path.joinpath("cfg2.xml")

    tf_cls1 = process_mth5(cfg1, kernel_dataset)
    tf_cls1.write(fn=cfg1_path, file_type="emtfxml")
    tf_cls2 = process_mth5(cfg2, kernel_dataset)
    tf_cls2.write(fn=cfg2_path, file_type="emtfxml")
    assert tfs_nearly_equal(tf_cls2, tf_cls1)



import pytest
from loguru import logger
from mth5.helpers import close_open_files
from mth5.processing import KernelDataset, RunSummary
from mth5.timeseries.spectre.helpers import (
    add_fcs_to_mth5,
    fc_decimations_creator,
    read_back_fcs,
)

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.process_mth5 import process_mth5
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.test_utils.synthetic.make_processing_configs import (
    create_test_run_config,
    make_processing_config_and_kernel_dataset,
)
from aurora.test_utils.synthetic.processing_helpers import process_synthetic_2
from aurora.test_utils.synthetic.triage import tfs_nearly_equal


@pytest.fixture(scope="module")
def mth5_test_files(
    worker_safe_test1_h5,
    worker_safe_test2_h5,
    worker_safe_test3_h5,
    worker_safe_test12rr_h5,
):
    """Create synthetic MTH5 test files."""
    logger.info("Making synthetic data")
    close_open_files()

    return {
        "paths": [
            worker_safe_test1_h5,
            worker_safe_test2_h5,
            worker_safe_test3_h5,
            worker_safe_test12rr_h5,
        ],
        "path_2": worker_safe_test2_h5,
    }


@pytest.mark.parametrize(
    "mth5_fixture_name",
    [
        "worker_safe_test1_h5",
        "worker_safe_test2_h5",
        "worker_safe_test3_h5",
        "worker_safe_test12rr_h5",
    ],
)
def test_add_fcs_to_synthetic_file(mth5_fixture_name, request, subtests):
    """Test adding Fourier Coefficients to a synthetic file.

    Uses the to_fc_decimation() method of AuroraDecimationLevel.
    Tests each step of the workflow with detailed validation:
    1. File validation (exists, can open, has structure)
    2. RunSummary creation and validation
    3. KernelDataset creation and validation
    4. Processing config creation and validation
    5. FC addition and validation
    6. FC readback validation
    7. Processing with FCs

    This test is parameterized to run separately for each MTH5 file,
    allowing parallel execution across different workers.
    """
    from mth5 import mth5

    # Get the actual fixture value using request.getfixturevalue
    mth5_path = request.getfixturevalue(mth5_fixture_name)
    subtest_name = mth5_path.stem

    logger.info(f"\n{'='*80}\nTesting {mth5_path.stem}\n{'='*80}")

    # Step 1: File validation
    with subtests.test(step=f"{subtest_name}_file_exists"):
        assert mth5_path.exists(), f"{mth5_path.stem} not found at {mth5_path}"
        logger.info(f"✓ File exists: {mth5_path}")

    with subtests.test(step=f"{subtest_name}_file_opens"):
        with mth5.MTH5(file_version="0.1.0") as m:
            m.open_mth5(mth5_path, mode="r")
            stations = m.stations_group.groups_list
            assert len(stations) > 0, f"No stations found in {mth5_path.stem}"
            logger.info(f"✓ File opens, stations: {stations}")

    with subtests.test(step=f"{subtest_name}_has_runs_and_channels"):
        with mth5.MTH5(file_version="0.1.0") as m:
            m.open_mth5(mth5_path, mode="r")
            for station_id in m.stations_group.groups_list:
                station = m.get_station(station_id)
                runs = [
                    r
                    for r in station.groups_list
                    if r
                    not in [
                        "Transfer_Functions",
                        "Fourier_Coefficients",
                        "Features",
                    ]
                ]
                assert len(runs) > 0, f"Station {station_id} has no runs"

                for run_id in runs:
                    run = station.get_run(run_id)
                    channels = run.groups_list
                    assert len(channels) > 0, f"Run {run_id} has no channels"

                    # Verify channels have data
                    for ch_name in channels:
                        ch = run.get_channel(ch_name)
                        assert ch.n_samples > 0, f"Channel {ch_name} has no data"

                logger.info(
                    f"✓ Station {station_id}: {len(runs)} run(s), channels validated"
                )

    # Step 2: RunSummary creation and validation
    with subtests.test(step=f"{subtest_name}_run_summary"):
        mth5_paths = [mth5_path]
        run_summary = RunSummary()
        run_summary.from_mth5s(mth5_paths)

        assert len(run_summary.df) > 0, f"RunSummary is empty for {mth5_path.stem}"

        # Validate sample rates are positive
        invalid_rates = run_summary.df[run_summary.df.sample_rate <= 0]
        assert len(invalid_rates) == 0, (
            f"RunSummary has {len(invalid_rates)} entries with invalid sample_rate:\n"
            f"{invalid_rates[['station', 'run', 'sample_rate']]}"
        )

        logger.info(
            f"✓ RunSummary: {len(run_summary.df)} entries, "
            f"sample_rates={run_summary.df.sample_rate.unique()}"
        )

    # Step 3: KernelDataset creation and validation
    with subtests.test(step=f"{subtest_name}_kernel_dataset"):
        tfk_dataset = KernelDataset()

        # Get Processing Config - determine station IDs
        if mth5_path.stem in ["test1", "test2"]:
            station_id = mth5_path.stem
            tfk_dataset.from_run_summary(run_summary, station_id)
        elif mth5_path.stem in ["test3"]:
            station_id = "test3"
            tfk_dataset.from_run_summary(run_summary, station_id)
        elif mth5_path.stem in ["test12rr"]:
            tfk_dataset.from_run_summary(run_summary, "test1", "test2")

        assert len(tfk_dataset.df) > 0, f"KernelDataset is empty for {mth5_path.stem}"
        assert (
            "station" in tfk_dataset.df.columns
        ), "KernelDataset missing 'station' column"
        assert "run" in tfk_dataset.df.columns, "KernelDataset missing 'run' column"

        logger.info(
            f"✓ KernelDataset: {len(tfk_dataset.df)} entries, "
            f"stations={tfk_dataset.df.station.unique()}"
        )

    # Step 4: Processing config creation and validation
    with subtests.test(step=f"{subtest_name}_processing_config"):
        if mth5_path.stem in ["test1", "test2"]:
            processing_config = create_test_run_config(station_id, tfk_dataset)
        elif mth5_path.stem in ["test3", "test12rr"]:
            cc = ConfigCreator()
            processing_config = cc.create_from_kernel_dataset(tfk_dataset)

        assert processing_config is not None, "Processing config is None"
        assert (
            len(processing_config.decimations) > 0
        ), "No decimations in processing config"
        assert (
            processing_config.channel_nomenclature is not None
        ), "No channel nomenclature"

        logger.info(
            f"✓ Processing config: {len(processing_config.decimations)} decimations"
        )

    # Step 5: FC addition and validation
    with subtests.test(step=f"{subtest_name}_add_fcs"):
        # Extract FC decimations from processing config
        fc_decimations = [x.to_fc_decimation() for x in processing_config.decimations]
        # For code coverage, test with fc_decimations=None for test1
        if mth5_path.stem == "test1":
            fc_decimations = None

        # Verify no FC group before adding
        with mth5.MTH5(file_version="0.1.0") as m:
            m.open_mth5(mth5_path, mode="r")
            for station_id in m.stations_group.groups_list:
                station = m.get_station(station_id)
                groups_before = station.groups_list
                # FC group might already exist from previous runs, but should be empty or absent

        add_fcs_to_mth5(mth5_path, fc_decimations=fc_decimations)

        # Validate FC group exists and has content
        with mth5.MTH5(file_version="0.1.0") as m:
            m.open_mth5(mth5_path, mode="r")
            for station_id in m.stations_group.groups_list:
                station = m.get_station(station_id)
                groups_after = station.groups_list

                assert "Fourier_Coefficients" in groups_after, (
                    f"Fourier_Coefficients group not found in station {station_id} "
                    f"after adding FCs. Groups: {groups_after}"
                )

                fc_group = station.fourier_coefficients_group
                fc_runs = fc_group.groups_list
                assert (
                    len(fc_runs) > 0
                ), f"No FC runs found in station {station_id} after adding FCs"

                # Validate each FC run has decimation levels
                for fc_run_id in fc_runs:
                    fc_run = fc_group.get_fc_group(fc_run_id)
                    dec_levels = fc_run.groups_list
                    assert (
                        len(dec_levels) > 0
                    ), f"No decimation levels in FC run {fc_run_id}"

                logger.info(
                    f"✓ FCs added to station {station_id}: "
                    f"{len(fc_runs)} run(s), {len(dec_levels)} decimation level(s)"
                )

    # Step 6: FC readback validation
    with subtests.test(step=f"{subtest_name}_read_back_fcs"):
        # This tests that FCs can be read back from the file
        read_back_fcs(mth5_path)
        logger.info(f"✓ FCs read back successfully")

    # Step 7: Processing with FCs
    with subtests.test(step=f"{subtest_name}_process_with_fcs"):
        tfc = process_mth5(processing_config, tfk_dataset=tfk_dataset)

        assert tfc is not None, f"process_mth5 returned None for {mth5_path.stem}"
        assert hasattr(tfc, "station_metadata"), "TF object missing station_metadata"
        assert len(tfc.station_metadata.runs) > 0, "TF object has no runs in metadata"

        logger.info(
            f"✓ Processing completed: {type(tfc).__name__}, "
            f"{len(tfc.station_metadata.runs)} run(s) processed"
        )

    logger.info(f"✓ All tests passed for {mth5_path.stem}\n")


def test_fc_decimations_creator():
    """Test fc_decimations_creator utility function."""
    cfgs = fc_decimations_creator(initial_sample_rate=1.0)
    assert cfgs is not None

    # test time period must be of correct type
    with pytest.raises(NotImplementedError):
        time_period = ["2023-01-01T17:48:29", "2023-01-09T08:54:08"]
        fc_decimations_creator(1.0, time_period=time_period)


def test_create_then_use_stored_fcs_for_processing(
    mth5_test_files, synthetic_test_paths
):
    """Test creating and using stored Fourier Coefficients for processing."""
    AURORA_RESULTS_PATH = synthetic_test_paths.aurora_results_path
    mth5_path_2 = mth5_test_files["path_2"]

    z_file_path_1 = AURORA_RESULTS_PATH.joinpath("test2.zss")
    z_file_path_2 = AURORA_RESULTS_PATH.joinpath("test2_from_stored_fc.zss")
    tf1 = process_synthetic_2(
        force_make_mth5=True,
        z_file_path=z_file_path_1,
        save_fc=True,
        mth5_path=mth5_path_2,
    )
    tfk_dataset, processing_config = make_processing_config_and_kernel_dataset(
        config_keyword="test2",
        station_id="test2",
        remote_id=None,
        mth5s=[mth5_path_2],
        channel_nomenclature="default",
    )

    # Initialize a TF kernel to check for FCs
    original_window = processing_config.decimations[0].stft.window.type

    tfk = TransferFunctionKernel(dataset=tfk_dataset, config=processing_config)
    tfk.update_processing_summary()
    tfk.check_if_fcs_already_exist()
    assert (
        tfk.dataset_df.fc.all()
    )  # assert fcs True in dataframe -- i.e. they were detected.

    # now change the window type and show that FCs are not detected
    for decimation in processing_config.decimations:
        decimation.stft.window.type = "hamming"
    tfk = TransferFunctionKernel(dataset=tfk_dataset, config=processing_config)
    tfk.update_processing_summary()
    tfk.check_if_fcs_already_exist()
    assert not (
        tfk.dataset_df.fc.all()
    )  # assert fcs False in dataframe -- i.e. they were detected.

    # Now reprocess with the FCs
    for decimation in processing_config.decimations:
        decimation.stft.window.type = original_window
    tfk = TransferFunctionKernel(dataset=tfk_dataset, config=processing_config)
    tfk.update_processing_summary()
    tfk.check_if_fcs_already_exist()
    assert (
        tfk.dataset_df.fc.all()
    )  # assert fcs True in dataframe -- i.e. they were detected.

    tf2 = process_synthetic_2(
        force_make_mth5=False, z_file_path=z_file_path_2, mth5_path=mth5_path_2
    )
    assert tfs_nearly_equal(tf1, tf2)



"""Pytest translation of `test_transfer_function_kernel.py`.

Uses fixtures and subtests. Designed to be xdist-safe by avoiding global
state and using fixtures from `conftest.py` where appropriate.
"""

from __future__ import annotations

import pytest

from aurora.config.config_creator import ConfigCreator
from aurora.pipelines.transfer_function_kernel import TransferFunctionKernel
from aurora.test_utils.synthetic.processing_helpers import get_example_kernel_dataset


@pytest.fixture
def kernel_dataset():
    return get_example_kernel_dataset()


@pytest.fixture
def processing_config(kernel_dataset):
    cc = ConfigCreator()
    return cc.create_from_kernel_dataset(kernel_dataset, estimator={"engine": "RME"})


@pytest.fixture
def tfk(kernel_dataset, processing_config):
    return TransferFunctionKernel(dataset=kernel_dataset, config=processing_config)


def test_init(tfk):
    """Constructing a TransferFunctionKernel with a valid config succeeds."""
    assert isinstance(tfk, TransferFunctionKernel)


def test_cannot_init_without_processing_config():
    """Calling constructor with no args raises TypeError (same as original)."""
    with pytest.raises(TypeError):
        TransferFunctionKernel()


def test_tfk_basic_properties(tfk, subtests):
    """A few lightweight sanity checks using subtests for clearer output."""
    with subtests.test("has_dataset"):
        assert getattr(tfk, "dataset", None) is not None

    with subtests.test("has_config"):
        assert getattr(tfk, "config", None) is not None

    with subtests.test("string_repr"):
        # Ensure a simple repr/str path doesn't error; not asserting exact
        # content since it may change between implementations.
        s = str(tfk)
        assert isinstance(s, str)



"""
This module contains a class that holds parameters to control the iterations of
robust regression including convergence criteria and thresholds.

Based on Gary's IterControl.m in
iris_mt_scratch/egbert_codes-20210121T193218Z-001/egbert_codes/matlabPrototype_10-13-20/TF/classes
"""
import numpy as np
from loguru import logger

from aurora.transfer_function.regression.helper_functions import rme_beta


class IterControl(object):
    """
    Notes:

    Here is a class to hold variables that are used to control the regression:

    - currently this is used for variations on RME (Robust M-Estimator)
    - TODO: in the original matlab code there was a class property called `epsilon`, but it was unused; There was no documentation about its purpose, except possibly that the abstract base class solved Y = X*b + epsilon for b, complex-valued. Perhaps this was intended as an intrinsic tolerated noise level. The value of epsilon was set to 1000.
    - TODO: The return covariance boolean just initializes arrays of zeros. Needs to be made functional or removed.
    """

    def __init__(
        self,
        max_number_of_iterations: int = 10,
        max_number_of_redescending_iterations: int = 2,
        r0: float = 1.4,
        u0: float = 2.8,
        tolerance: float = 0.0005,
        verbosity: int = 0,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        max_number_of_iterations: int
            Set to zero for OLS, otherwise, this is how many times the RME
            will refine the estimate.
        max_number_of_redescending_iterations : int
            1 or 2 is fine at most.  If set to zero, the redescend code block is ignored.
        r0: float
            Effectively infinty for OLS, this controls the point at which residuals
            transition from being penalized by a squared vs a linear function.  The
            units are standard deviations of the residual distribution.  Normally
            1.4, 1.5 or thereabouts
        u0: float
            u0 is a parameter for the redescending portion of the robust regression.
            It is controlled by a double exponential formula (REFERENCE NEEDED).  It
            makes for severe downweighting about u0.  The function is continuous
            "math friendly" (all derivates exist etc) so you can prove theorems about it
            etc.
        tolerance : float
            minimum fractional change in any term in b.  Any smaller change
            than this will trigger convergence
        verbosity: int
            Allows setting of custom logging messages in development.

        Development notes:
        There was originally a parameter called epsilon (float) but it was not used.

        """
        self._number_of_iterations = 0
        self._number_of_redescending_iterations = 0

        self.tolerance = tolerance
        self.max_number_of_iterations = max_number_of_iterations
        self.max_number_of_redescending_iterations = (
            max_number_of_redescending_iterations
        )

        # regression-M estimator params
        self.r0 = r0
        self.u0 = u0

        self.verbosity = verbosity
        # Additional properties
        # used sometimes to control one or another of the iterative algorithms
        # These were translated from the matlab code and may be moved in future
        self.return_covariance = True  # TODO: add functionality
        self.save_cleaned = False  # TODO: add functionality
        self.robust_diagonalize = False  # TODO: add functionality

    @property
    def number_of_iterations(self) -> int:
        return self._number_of_iterations

    # @number_of_iterations.setter
    # def number_of_iterations(self, value) -> int:
    #     self._number_of_iterations = value

    def reset_number_of_iterations(self) -> int:
        self._number_of_iterations = 0

    def increment_iteration_number(self):
        self._number_of_iterations += 1

    @property
    def number_of_redescending_iterations(self) -> int:
        return self._number_of_redescending_iterations

    def reset_number_of_redescending_iterations(self):
        self._number_of_redescending_iterations = 0

    def increment_redescending_iteration_number(self):
        self._number_of_redescending_iterations += 1

    @property
    def max_iterations_reached(self) -> bool:
        """
        Return True of the number of iterations carried out is greater or equal the
        maximum number of iterations set in the processing config
        """
        return self.number_of_iterations >= self.max_number_of_iterations

    def converged(self, b, b0):
        """
        Parameters
        ----------
        b : complex-valued numpy array
            the most recent regression estimate
        b0 : complex-valued numpy array
            The previous regression estimate

        Returns
        -------
        converged: bool
            True of the regression has terminated, False otherwise

        Developement Notes:
        The variable maximum_change finds the maximum amplitude component of the vector
        1 - b/b0.  Looking at the formula, one might want to cast this instead as
        1 - abs(b/b0), however, that will be insensitive to phase changes in b,
        which is complex valued.  The way it is coded np.max(np.abs(1 - b / b0)) is
        correct as it stands.

        """

        maximum_change = np.max(np.abs(1 - b / b0))
        tolerance_cond = maximum_change <= self.tolerance
        iteration_cond = self.max_iterations_reached
        if tolerance_cond or iteration_cond:
            converged = True
            if self.verbosity > 0:
                msg_start = "Converged due to"
                msg_end = (
                    f"{self.number_of_iterations} of "
                    f"{self.max_number_of_iterations} iterations"
                )
                if tolerance_cond:
                    msg = f"{msg_start} MaxChange < Tolerance after {msg_end}"
                elif iteration_cond:
                    msg = f"{msg_start} maximum number_of_iterations {msg_end}"
                logger.info(msg)
        else:
            converged = False

        return converged

    @property
    def continue_redescending(self) -> bool:
        """
            Checks if the max_number_of_redescending_iterations has been reached

        Returns
        -------
        maxxed_out: bool
            True if max_number_of_redescending_iterations has been reached, otherwise False.

        """
        maxxed_out = (
            self.number_of_redescending_iterations
            >= self.max_number_of_redescending_iterations
        )
        if maxxed_out:
            return False
        else:
            return True

    @property
    def correction_factor(self) -> float:
        """
        Returns correction factor for residual variances.

        Detailed notes on usage in
        transfer_function.regression.helper_functions.rme_beta

        TODO: This is an RME specific property.  Suggest move r0, u0 and this method
         into an RME-config class.

        Returns
        -------
        correction_factor : float
            correction factor used for scaling the residual error_variance
        """
        correction_factor = 1.0 / rme_beta(self.r0)
        return correction_factor



"""Minimal conftest for aurora tests that need small mth5 fixtures.

This provides a small, self-contained subset of the mth5 test fixtures
so aurora tests can create and use `test12rr` MTH5 files without depending
on the mth5 repo's conftest discovery.

Fixtures provided:
- `worker_id` : pytest-xdist aware worker id
- `make_worker_safe_path(base, directory)` : make worker-unique filenames
- `fresh_test12rr_mth5` : creates a fresh `test12rr` MTH5 file in `tmp_path`
- `cleanup_test_files` : register files to be removed at session end
"""

# Set non-interactive matplotlib backend before any other imports
# This prevents tests from blocking on figure windows
import matplotlib


matplotlib.use("Agg")

from pathlib import Path
from typing import Dict

import pytest
from mt_metadata.transfer_functions.core import TF as _MT_TF
from mth5.data.make_mth5_from_asc import (
    create_test1_h5,
    create_test2_h5,
    create_test3_h5,
    create_test12rr_h5,
)
from mth5.helpers import close_open_files

from aurora.test_utils.synthetic.paths import SyntheticTestPaths


# Monkeypatch TF.write to sanitize None provenance/comment fields that cause
# pydantic validation errors when writing certain formats (e.g., emtfxml).
_orig_tf_write = getattr(_MT_TF, "write", None)


def _safe_tf_write(self, *args, **kwargs):
    # Pre-emptively sanitize station provenance comments to avoid pydantic errors
    try:
        sm = getattr(self, "station_metadata", None)
        if sm is not None:
            # Handle dict-based metadata (from pydantic branch)
            if isinstance(sm, dict):
                prov = sm.get("provenance")
                if prov and isinstance(prov, dict):
                    archive = prov.get("archive")
                    if archive and isinstance(archive, dict):
                        comments = archive.get("comments")
                        if comments and isinstance(comments, dict):
                            if comments.get("value") is None:
                                comments["value"] = ""
            else:
                # Handle object-based metadata (traditional approach)
                sm_list = (
                    sm if hasattr(sm, "__iter__") and not isinstance(sm, str) else [sm]
                )
                for s in sm_list:
                    try:
                        prov = getattr(s, "provenance", None)
                        if prov is None:
                            continue
                        archive = getattr(prov, "archive", None)
                        if archive is None:
                            continue
                        comments = getattr(archive, "comments", None)
                        if comments is None:
                            from types import SimpleNamespace

                            archive.comments = SimpleNamespace(value="")
                        elif getattr(comments, "value", None) is None:
                            comments.value = ""
                    except Exception:
                        pass
    except Exception:
        pass
    # Call original write
    return _orig_tf_write(self, *args, **kwargs)


if _orig_tf_write is not None:
    setattr(_MT_TF, "write", _safe_tf_write)


# Suppress noisy third-party deprecation and pydantic serializer warnings
# that are not actionable in these tests. These originate from external
# dependencies (jupyter_client, obspy/pkg_resources) and from pydantic when
# receiving plain strings where enums are expected. Filtering here keeps test
# output focused on real failures.
# warnings.filterwarnings(
#     "ignore",
#     category=UserWarning,
#     message=r"Pydantic serializer warnings:.*",
# )
# warnings.filterwarnings(
#     "ignore",
#     category=DeprecationWarning,
#     message=r"Jupyter is migrating its paths to use standard platformdirs",
# )
# warnings.filterwarnings(
#     "ignore",
#     category=DeprecationWarning,
#     message=r"pkg_resources",
# )
# warnings.filterwarnings(
#     "ignore",
#     category=DeprecationWarning,
#     message=r"np\.bool",
# )


# Process-wide cache for heavyweight test artifacts (keyed by worker id)
# stores the created MTH5 file path so multiple tests in the same session
# / worker can reuse the same file rather than recreating it repeatedly.
_MTH5_GLOBAL_CACHE: Dict[str, str] = {}


@pytest.fixture(scope="session")
def worker_id(request):
    """Return pytest-xdist worker id or 'master' when not using xdist."""
    if hasattr(request.config, "workerinput"):
        return request.config.workerinput.get("workerid", "gw0")
    return "master"


def get_worker_safe_filename(base_filename: str, worker: str) -> str:
    p = Path(base_filename)
    return f"{p.stem}_{worker}{p.suffix}"


@pytest.fixture
def make_worker_safe_path(worker_id):
    """Factory to produce worker-safe paths.

    Usage: `p = make_worker_safe_path('name.zrr', tmp_path)`
    """

    def _make(base_filename: str, directory: Path | None = None) -> Path:
        safe_name = get_worker_safe_filename(base_filename, worker_id)
        if directory is None:
            return Path(safe_name)
        return Path(directory) / safe_name

    return _make


@pytest.fixture(scope="session")
def synthetic_test_paths(tmp_path_factory, worker_id):
    """Create a SyntheticTestPaths instance that writes into a worker-unique tmp sandbox.

    This keeps tests isolated across xdist workers and avoids writing into the repo.
    """
    base = tmp_path_factory.mktemp(f"synthetic_{worker_id}")
    stp = SyntheticTestPaths(sandbox_path=base)
    stp.mkdirs()
    return stp


@pytest.fixture(autouse=True)
def ensure_closed_files():
    """Ensure mth5 open files are closed before/after each test to avoid cross-test leaks."""
    # run before test
    close_open_files()
    yield
    # run after test
    close_open_files()


@pytest.fixture(scope="session")
def cleanup_test_files(request):
    files = []

    def _register(p: Path):
        if p not in files:
            files.append(p)

    def _cleanup():
        for p in files:
            try:
                if p.exists():
                    p.unlink()
            except Exception:
                # best-effort cleanup
                pass

    request.addfinalizer(_cleanup)
    return _register


@pytest.fixture(scope="session")
def fresh_test12rr_mth5(mth5_target_dir: Path, worker_id, cleanup_test_files):
    """Create a fresh `test12rr` MTH5 file in mth5_target_dir and return its Path.

    This is intentionally simple: it calls `create_test12rr_h5` with a
    temporary target folder. The resulting file is registered for cleanup.
    Session-scoped for efficiency.
    """
    cache_key = f"test12rr_{worker_id}"

    # Return cached file if present and still exists
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # create_test12rr_h5 returns the path to the file it created
    # Use the session-scoped mth5_target_dir
    file_path = create_test12rr_h5(target_folder=mth5_target_dir)

    # register cleanup and cache
    ppath = Path(file_path)
    cleanup_test_files(ppath)
    _MTH5_GLOBAL_CACHE[cache_key] = str(ppath)

    return ppath


@pytest.fixture(scope="session")
def mth5_target_dir(tmp_path_factory, worker_id):
    """Create a worker-safe directory for MTH5 file creation.

    This directory is shared across all tests in a worker session,
    allowing MTH5 files to be cached and reused within a worker.
    """
    base_dir = tmp_path_factory.mktemp(f"mth5_files_{worker_id}")
    return base_dir


def _create_worker_safe_mth5(
    mth5_name: str,
    create_func,
    target_dir: Path,
    worker_id: str,
    file_version: str = "0.1.0",
    channel_nomenclature: str = "default",
    **kwargs,
) -> Path:
    """Helper to create worker-safe MTH5 files with caching.

    Parameters
    ----------
    mth5_name : str
        Base name for the MTH5 file (e.g., "test1", "test2")
    create_func : callable
        Function to create the MTH5 file (e.g., create_test1_h5)
    target_dir : Path
        Directory where the MTH5 file should be created
    worker_id : str
        Worker ID for pytest-xdist
    file_version : str
        MTH5 file version
    channel_nomenclature : str
        Channel nomenclature to use
    **kwargs
        Additional arguments to pass to create_func

    Returns
    -------
    Path
        Path to the created MTH5 file
    """
    cache_key = f"{mth5_name}_{worker_id}_{file_version}_{channel_nomenclature}"

    # Return cached file if present and still exists
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # Create the MTH5 file in the worker-safe directory
    file_path = create_func(
        file_version=file_version,
        channel_nomenclature=channel_nomenclature,
        target_folder=target_dir,
        force_make_mth5=True,
        **kwargs,
    )

    # Cache the path
    ppath = Path(file_path)
    _MTH5_GLOBAL_CACHE[cache_key] = str(ppath)

    return ppath


@pytest.fixture(scope="session")
def worker_safe_test1_h5(mth5_target_dir, worker_id):
    """Create test1.h5 in a worker-safe directory."""
    return _create_worker_safe_mth5(
        "test1", create_test1_h5, mth5_target_dir, worker_id
    )


@pytest.fixture(scope="session")
def worker_safe_test2_h5(mth5_target_dir, worker_id):
    """Create test2.h5 in a worker-safe directory."""
    return _create_worker_safe_mth5(
        "test2", create_test2_h5, mth5_target_dir, worker_id
    )


@pytest.fixture(scope="session")
def worker_safe_test3_h5(mth5_target_dir, worker_id):
    """Create test3.h5 in a worker-safe directory."""
    return _create_worker_safe_mth5(
        "test3", create_test3_h5, mth5_target_dir, worker_id
    )


@pytest.fixture(scope="session")
def worker_safe_test12rr_h5(mth5_target_dir, worker_id):
    """Create test12rr.h5 in a worker-safe directory."""
    return _create_worker_safe_mth5(
        "test12rr", create_test12rr_h5, mth5_target_dir, worker_id
    )


# ============================================================================
# Parkfield Test Fixtures
# ============================================================================


@pytest.fixture(scope="session")
def parkfield_paths():
    """Provide Parkfield test data paths."""
    from aurora.test_utils.parkfield.path_helpers import PARKFIELD_PATHS

    return PARKFIELD_PATHS


@pytest.fixture(scope="session")
def parkfield_h5_master(tmp_path_factory):
    """Create the master Parkfield MTH5 file once per test session.

    This downloads data from NCEDC and caches it in a persistent directory
    (.cache/aurora/parkfield) so it doesn't need to be re-downloaded for
    subsequent test runs. Only created once across all sessions.
    """
    from filelock import FileLock

    from aurora.test_utils.parkfield.make_parkfield_mth5 import ensure_h5_exists

    # Use a persistent cache directory instead of temp
    # This way the file survives across test sessions
    cache_dir = Path.home() / ".cache" / "aurora" / "parkfield"
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Check if file already exists in persistent cache
    cached_file = cache_dir / "parkfield.h5"
    lock_file = cache_dir / "parkfield.h5.lock"

    # Check global cache first (for current session)
    cache_key = "parkfield_master"
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # Quick check before acquiring lock - avoid contention if file exists
    if cached_file.exists():
        _MTH5_GLOBAL_CACHE[cache_key] = str(cached_file)
        return cached_file

    # Use filelock to ensure only one worker creates the file
    with FileLock(str(lock_file), timeout=300):
        # Double-check after acquiring lock (another worker may have created it)
        if cached_file.exists():
            _MTH5_GLOBAL_CACHE[cache_key] = str(cached_file)
            return cached_file

        try:
            h5_path = ensure_h5_exists(target_folder=cache_dir)
            _MTH5_GLOBAL_CACHE[cache_key] = str(h5_path)
            return h5_path
        except IOError:
            pytest.skip("NCEDC data server not available")


@pytest.fixture(scope="session")
def parkfield_h5_path(parkfield_h5_master, tmp_path_factory, worker_id):
    """Copy master Parkfield MTH5 to worker-safe location.

    The master file is created once and cached persistently in
    ~/.cache/aurora/parkfield/ so it doesn't need to be re-downloaded.
    This fixture copies that cached file to a worker-specific temp
    directory to avoid file handle conflicts in pytest-xdist parallel execution.
    """
    import shutil

    cache_key = f"parkfield_h5_{worker_id}"

    # Check cache first
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # Create worker-safe directory and copy the master file
    target_dir = tmp_path_factory.mktemp(f"parkfield_{worker_id}")
    worker_h5_path = target_dir / parkfield_h5_master.name

    shutil.copy2(parkfield_h5_master, worker_h5_path)
    _MTH5_GLOBAL_CACHE[cache_key] = str(worker_h5_path)
    return worker_h5_path


@pytest.fixture
def parkfield_mth5(parkfield_h5_path):
    """Open and close MTH5 object for Parkfield data.

    This is a function-scoped fixture that ensures proper cleanup
    of MTH5 file handles after each test.
    """
    from mth5.mth5 import MTH5

    mth5_obj = MTH5(file_version="0.1.0")
    mth5_obj.open_mth5(parkfield_h5_path, mode="r")
    yield mth5_obj
    mth5_obj.close_mth5()


@pytest.fixture
def parkfield_run_pkd(parkfield_mth5):
    """Get PKD station run 001 from Parkfield MTH5."""
    run_obj = parkfield_mth5.get_run("PKD", "001")
    return run_obj


@pytest.fixture
def parkfield_run_ts_pkd(parkfield_run_pkd):
    """Get RunTS object for PKD station."""
    return parkfield_run_pkd.to_runts()


@pytest.fixture(scope="class")
def parkfield_kernel_dataset_ss(parkfield_h5_path):
    """Create single-station KernelDataset for PKD."""
    from mth5.processing import KernelDataset, RunSummary

    run_summary = RunSummary()
    run_summary.from_mth5s([parkfield_h5_path])
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(run_summary=run_summary, local_station_id="PKD")
    return tfk_dataset


@pytest.fixture(scope="class")
def parkfield_kernel_dataset_rr(parkfield_h5_path):
    """Create remote-reference KernelDataset for PKD with SAO as RR."""
    from mth5.processing import KernelDataset, RunSummary

    run_summary = RunSummary()
    run_summary.from_mth5s([parkfield_h5_path])
    tfk_dataset = KernelDataset()
    tfk_dataset.from_run_summary(
        run_summary=run_summary, local_station_id="PKD", remote_station_id="SAO"
    )
    return tfk_dataset


@pytest.fixture
def disable_matplotlib_logging(request):
    """Disable noisy matplotlib logging for cleaner test output."""
    import logging

    loggers_to_disable = [
        "matplotlib.font_manager",
        "matplotlib.ticker",
    ]

    original_states = {}
    for logger_name in loggers_to_disable:
        logger_obj = logging.getLogger(logger_name)
        original_states[logger_name] = logger_obj.disabled
        logger_obj.disabled = True

    yield

    # Restore original states
    for logger_name, original_state in original_states.items():
        logging.getLogger(logger_name).disabled = original_state


# =============================================================================
# CAS04 FDSN Fixtures
# =============================================================================


@pytest.fixture(scope="session")
def _master_fdsn_miniseed_v010():
    """Master CAS04 FDSN MTH5 file (v0.1.0) - created once, copied per worker.

    Uses persistent cache in ~/.cache/aurora/cas04/ to avoid recreating
    the file across test sessions and CI runs.
    """
    import obspy
    from filelock import FileLock
    from mth5.clients.fdsn import FDSN
    from mth5_test_data import get_test_data_path

    # Use a persistent cache directory instead of temp
    cache_dir = Path.home() / ".cache" / "aurora" / "cas04"
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Check if file already exists in persistent cache
    master_file = cache_dir / "cas04_v010_master.h5"
    lock_file = cache_dir / "cas04_v010_master.h5.lock"

    # Quick check before acquiring lock - avoid contention if file exists
    if master_file.exists():
        return master_file

    # Use filelock to ensure only one worker creates the file
    with FileLock(str(lock_file), timeout=300):
        # Double-check after acquiring lock (another worker may have created it)
        if master_file.exists():
            return master_file

        # Get test data paths
        miniseed_path = get_test_data_path("miniseed")
        inventory_file = miniseed_path / "cas04_stationxml.xml"
        streams_file = miniseed_path / "cas_04_streams.mseed"

        # Verify files exist
        if not inventory_file.exists() or not streams_file.exists():
            pytest.skip(
                f"CAS04 test data not found in mth5_test_data. Expected:\n"
                f"  {inventory_file}\n"
                f"  {streams_file}"
            )

        # Load inventory and streams
        inventory = obspy.read_inventory(str(inventory_file))
        streams = obspy.read(str(streams_file))

        # Create MTH5 from inventory and streams in cache directory
        # The function creates a file with a default name, we need to rename it
        fdsn_client = FDSN(mth5_version="0.1.0")
        created_file = fdsn_client.make_mth5_from_inventory_and_streams(
            inventory, streams, save_path=cache_dir
        )

        # Rename to version-specific master file
        created_path = Path(created_file)
        if created_path != master_file:
            import shutil

            shutil.move(str(created_path), str(master_file))

    return master_file


@pytest.fixture(scope="session")
def global_fdsn_miniseed_v010(_master_fdsn_miniseed_v010, mth5_target_dir, worker_id):
    """Worker-safe copy of CAS04 v0.1.0 MTH5 file for parallel testing.

    Creates a per-worker copy of the master file to avoid concurrent access issues.
    """
    import shutil

    # Check worker-specific cache
    cache_key = f"cas04_v010_{worker_id}"
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # Copy master file to worker-specific location
    worker_file = mth5_target_dir / f"cas04_v010_{worker_id}.h5"
    shutil.copy2(_master_fdsn_miniseed_v010, worker_file)

    # Cache the worker-specific path
    _MTH5_GLOBAL_CACHE[cache_key] = str(worker_file)

    return worker_file


@pytest.fixture(scope="session")
def _master_fdsn_miniseed_v020():
    """Master CAS04 FDSN MTH5 file (v0.2.0) - created once, copied per worker.

    Uses persistent cache in ~/.cache/aurora/cas04/ to avoid recreating
    the file across test sessions and CI runs.
    """
    import obspy
    from filelock import FileLock
    from mth5.clients.fdsn import FDSN
    from mth5_test_data import get_test_data_path

    # Use a persistent cache directory instead of temp
    cache_dir = Path.home() / ".cache" / "aurora" / "cas04"
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Check if file already exists in persistent cache
    master_file = cache_dir / "cas04_v020_master.h5"
    lock_file = cache_dir / "cas04_v020_master.h5.lock"

    # Quick check before acquiring lock - avoid contention if file exists
    if master_file.exists():
        return master_file

    # Use filelock to ensure only one worker creates the file
    with FileLock(str(lock_file), timeout=300):
        # Double-check after acquiring lock (another worker may have created it)
        if master_file.exists():
            return master_file

        # Get test data paths
        miniseed_path = get_test_data_path("miniseed")
        inventory_file = miniseed_path / "cas04_stationxml.xml"
        streams_file = miniseed_path / "cas_04_streams.mseed"

        # Verify files exist
        if not inventory_file.exists() or not streams_file.exists():
            pytest.skip(
                f"CAS04 test data not found in mth5_test_data. Expected:\n"
                f"  {inventory_file}\n"
                f"  {streams_file}"
            )

        # Load inventory and streams
        inventory = obspy.read_inventory(str(inventory_file))
        streams = obspy.read(str(streams_file))

        # Create MTH5 from inventory and streams in cache directory
        # The function creates a file with a default name, we need to rename it
        fdsn_client = FDSN(mth5_version="0.2.0")
        created_file = fdsn_client.make_mth5_from_inventory_and_streams(
            inventory, streams, save_path=cache_dir
        )

        # Rename to version-specific master file
        created_path = Path(created_file)
        if created_path != master_file:
            import shutil

            shutil.move(str(created_path), str(master_file))

    return master_file


@pytest.fixture(scope="session")
def global_fdsn_miniseed_v020(_master_fdsn_miniseed_v020, mth5_target_dir, worker_id):
    """Worker-safe copy of CAS04 v0.2.0 MTH5 file for parallel testing.

    Creates a per-worker copy of the master file to avoid concurrent access issues.
    """
    import shutil

    # Check worker-specific cache
    cache_key = f"cas04_v020_{worker_id}"
    cached = _MTH5_GLOBAL_CACHE.get(cache_key)
    if cached:
        p = Path(cached)
        if p.exists():
            return p

    # Copy master file to worker-specific location
    worker_file = mth5_target_dir / f"cas04_v020_{worker_id}.h5"
    shutil.copy2(_master_fdsn_miniseed_v020, worker_file)

    # Cache the worker-specific path
    _MTH5_GLOBAL_CACHE[cache_key] = str(worker_file)

    return worker_file
